import enum
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang


##


class QueryMode(enum.Enum):
    QUERY = enum.auto()
    EXEC = enum.auto()

    @classmethod
    def of(
            cls,
            o: ta.Union[str, 'QueryMode', None],
            default: ta.Optional['QueryMode'] = None,
    ) -> 'QueryMode':
        if o is None:
            return check.not_none(check.isinstance(default, cls))
        elif isinstance(o, str):
            return cls[o.upper()]  # noqa
        elif isinstance(o, cls):
            return o
        else:
            raise TypeError(o)


##


class Queryable(lang.Sealed, lang.Abstract):
    pass


@ta.final
@dc.dataclass(frozen=True)
class Query(Queryable, lang.Final):
    mode: QueryMode
    text: str
    args: ta.Sequence[ta.Any]



# @omlish-lite
# @omlish-amalg ../../../omdev/scripts/lib/io/streams.py
from .adapters import BytesIoByteStreamBuffer  # noqa
from .direct import DirectByteStreamBuffer  # noqa
from .framing import LongestMatchDelimiterByteStreamFrameDecoder  # noqa
from .linear import LinearByteStreamBuffer  # noqa
from .reading import ByteStreamBufferReader  # noqa
from .scanning import ScanningByteStreamBuffer  # noqa
from .segmented import SegmentedByteStreamBuffer  # noqa
from .utils import ByteStreamBuffers  # noqa


##



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import typing as ta

from ....logs.base import Logger
from ....logs.levels import NamedLogLevel
from ....logs.modules import get_module_logger
from ..core import IoPipelineDirection
from ..core import IoPipelineDirectionOrDuplex
from ..core import IoPipelineHandler
from ..core import IoPipelineHandlerContext


log = get_module_logger(globals())  # noqa


##


class LoggingIoPipelineHandler(IoPipelineHandler):
    def __init__(
            self,
            log: ta.Optional[Logger] = None,  # noqa
            level: ta.Union[int, str] = 'debug',
            *,
            direction: IoPipelineDirectionOrDuplex = 'duplex',
    ) -> None:
        super().__init__()

        if log is None:
            log = globals()['log']  # noqa
        self._log = log
        self._level = NamedLogLevel(level)
        self._direction = direction

    def _do_log(
            self,
            ctx: IoPipelineHandlerContext,
            direction: IoPipelineDirection,
            msg: ta.Any,
    ) -> None:
        self._log.log(
            self._level,
            '%s: %r',
            direction,
            msg,
        )

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if self._direction in ('inbound', 'duplex'):
            self._do_log(ctx, 'inbound', msg)
        ctx.feed_in(msg)

    def outbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if self._direction in ('outbound', 'duplex'):
            self._do_log(ctx, 'outbound', msg)
        ctx.feed_out(msg)



import abc

from ... import lang
from ..params import ParamStyle


##


class Adapter(lang.Abstract):
    @property
    @abc.abstractmethod
    def param_style(self) -> ParamStyle | None:
        raise NotImplementedError



"""
TODO:
 - https://github.com/mag37/dockcheck/blob/3d122f2b868eb53a25a3014f0f6bd499390a3a29/dockcheck.sh
  - https://github.com/regclient/regclient
  - https://stackoverflow.com/questions/71409458/how-to-download-docker-image-using-http-api-using-docker-hub-credentials
  - https://stackoverflow.com/questions/55386202/how-can-i-use-the-docker-registry-api-to-pull-information-about-a-container-get
  - https://ops.tips/blog/inspecting-docker-image-without-pull/
"""  # noqa
import datetime
import re
import subprocess
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .. import marshal as msh
from ..formats import json


##


@dc.dataclass(frozen=True)
@msh.update_object_options(field_naming=msh.Naming.CAMEL, unknown_field='x')
@msh.update_fields_options(['id'], name='ID')
class PsItem(lang.Final):
    command: str
    created_at: datetime.datetime
    id: str
    image: str
    labels: str
    local_volumes: str
    mounts: str
    names: str
    networks: str
    ports: str
    running_for: str
    size: str
    state: str
    status: str

    x: ta.Mapping[str, ta.Any] | None = None


class Port(ta.NamedTuple):
    ip: str
    from_port: int
    to_port: int
    proto: str


_PORT_PAT = re.compile(r'(?P<ip>[^:]+):(?P<from_port>\d+)->(?P<to_port>\d+)/(?P<proto>\w+)')


def parse_port(s: str) -> Port:
    # '0.0.0.0:35221->22/tcp, 0.0.0.0:35220->8000/tcp'
    m = check.not_none(_PORT_PAT.fullmatch(s))
    return Port(
        m.group('ip'),
        int(m.group('from_port')),
        int(m.group('to_port')),
        m.group('proto'),
    )


def cli_ps() -> list[PsItem]:
    o = subprocess.check_output([
        'docker',
        'ps',
        '--no-trunc',
        '--format', '{{json .}}',
    ])

    ret: list[PsItem] = []
    for l in o.decode().splitlines():
        d = json.loads(l)
        pi = msh.unmarshal(d, PsItem)
        ret.append(pi)

    return ret


@dc.dataclass(frozen=True)
@msh.update_object_options(field_naming=msh.Naming.CAMEL, unknown_field='x')
class Inspect(lang.Final):
    id: str
    created: datetime.datetime

    x: ta.Mapping[str, ta.Any] | None = None


def cli_inspect(ids: list[str]) -> list[Inspect]:
    o = subprocess.check_output(['docker', 'inspect', *ids])
    return msh.unmarshal(json.loads(o.decode()), list[Inspect])


def has_cli() -> bool:
    try:
        proc = subprocess.run(['docker', '--version'])  # noqa
    except (FileNotFoundError, subprocess.CalledProcessError):
        return False
    else:
        return not proc.returncode



"""
https://datatracker.ietf.org/doc/html/rfc5234
"""
import typing as ta

from .grammars import Channel
from .grammars import Grammar
from .grammars import Rule
from .ops import concat
from .ops import either
from .ops import literal
from .ops import repeat
from .ops import rule
from .opto import optimize_grammar


##


CORE_RULES: ta.Sequence[Rule] = [

    Rule(
        'ALPHA',
        either(
            literal('\x41', '\x5a'),
            literal('\x61', '\x7a'),
        ),
        channel=Channel.CONTENT,
    ),

    Rule(
        'BIT',
        either(
            literal('0'),
            literal('1'),
        ),
        channel=Channel.CONTENT,
    ),

    Rule(
        'CHAR',
        literal('\x01', '\x7f'),
        channel=Channel.CONTENT,
    ),

    Rule(
        'CTL',
        either(
            literal('\x00', '\x1f'),
            literal('\x7f', case_sensitive=True),
        ),
        channel=Channel.CONTENT,
    ),

    Rule(
        'CR',
        literal('\x0d', case_sensitive=True),
        channel=Channel.SPACE,
    ),

    Rule(
        'CRLF',
        concat(
            rule('CR'),
            rule('LF'),
        ),
        channel=Channel.SPACE,
    ),

    Rule(
        'DIGIT',
        literal('\x30', '\x39'),
        channel=Channel.CONTENT,
    ),

    Rule(
        'DQUOTE',
        literal('\x22', case_sensitive=True),
        channel=Channel.CONTENT,
    ),

    Rule(
        'HEXDIG',
        either(
            rule('DIGIT'),
            literal('A'),
            literal('B'),
            literal('C'),
            literal('D'),
            literal('E'),
            literal('F'),
        ),
        channel=Channel.CONTENT,
    ),

    Rule(
        'HTAB',
        literal('\x09', case_sensitive=True),
        channel=Channel.SPACE,
    ),

    Rule(
        'LF',
        literal('\x0a', case_sensitive=True),
        channel=Channel.SPACE,
    ),

    Rule(
        'LWSP',
        repeat(
            either(
                rule('WSP'),
                concat(
                    rule('CRLF'),
                    rule('WSP'),
                ),
            ),
        ),
        channel=Channel.SPACE,
    ),

    Rule(
        'OCTET',
        literal('\x00', '\xff'),
        channel=Channel.CONTENT,
    ),

    Rule(
        'SP',
        literal('\x20', case_sensitive=True),
        channel=Channel.SPACE,
    ),

    Rule(
        'VCHAR',
        literal('\x21', '\x7e'),
        channel=Channel.CONTENT,
    ),

    Rule(
        'WSP',
        either(
            rule('SP'),
            rule('HTAB'),
        ),
        channel=Channel.SPACE,
    ),

]


RAW_CORE_GRAMMAR = Grammar(*CORE_RULES)
CORE_GRAMMAR = optimize_grammar(RAW_CORE_GRAMMAR)



# ruff: noqa: UP006 UP007 UP045
import dataclasses as dc
import functools
import typing as ta


##


def dataclass_shallow_astuple(o: ta.Any) -> ta.Tuple[ta.Any, ...]:
    return tuple(getattr(o, f.name) for f in dc.fields(o))


def dataclass_shallow_asdict(o: ta.Any) -> ta.Dict[str, ta.Any]:
    return {f.name: getattr(o, f.name) for f in dc.fields(o)}


##


def is_immediate_dataclass(cls: type) -> bool:
    if not isinstance(cls, type):
        raise TypeError(cls)
    return dc._FIELDS in cls.__dict__  # type: ignore[attr-defined]  # noqa


##


def _install_dataclass_fn(cls: type, fn: ta.Any, fn_name: ta.Optional[str] = None) -> None:
    if fn_name is None:
        fn_name = fn.__name__
    setattr(cls, fn_name, fn)
    fn.__qualname__ = f'{cls.__qualname__}.{fn_name}'


##


def install_dataclass_cache_hash(
        *,
        cached_hash_attr: str = '__dataclass_hash__',
):
    def inner(cls):
        if not isinstance(cls, type) and dc.is_dataclass(cls):
            raise TypeError(cls)

        if (
                cls.__hash__ is object.__hash__ or
                '__hash__' not in cls.__dict__
        ):
            raise TypeError(cls)

        real_hash = cls.__hash__

        def cached_hash(self) -> int:
            try:
                return object.__getattribute__(self, cached_hash_attr)
            except AttributeError:
                object.__setattr__(self, cached_hash_attr, h := real_hash(self))  # type: ignore[call-arg]
            return h

        _install_dataclass_fn(cls, cached_hash, '__hash__')

        return cls

    return inner


##


def dataclass_maybe_post_init(sup: ta.Any) -> bool:
    if not isinstance(sup, super):
        raise TypeError(sup)
    try:
        fn = sup.__post_init__  # type: ignore
    except AttributeError:
        return False
    fn()
    return True


##


def dataclass_filtered_repr(
        obj: ta.Any,
        fn: ta.Union[ta.Callable[[ta.Any, dc.Field, ta.Any], bool], ta.Literal['omit_none', 'omit_falsey']],
) -> str:
    if fn == 'omit_none':
        fn = lambda o, f, v: v is not None  # noqa
    elif fn == 'omit_falsey':
        fn = lambda o, f, v: bool(v)

    return (
        f'{obj.__class__.__qualname__}(' +
        ', '.join([
            f'{f.name}={v!r}'
            for f in dc.fields(obj)
            if fn(obj, f, v := getattr(obj, f.name))
        ]) +
        ')'
    )


def dataclass_repr_omit_none(obj: ta.Any) -> str:
    return dataclass_filtered_repr(obj, 'omit_none')


def dataclass_repr_omit_falsey(obj: ta.Any) -> str:
    return dataclass_filtered_repr(obj, 'omit_falsey')


def install_dataclass_filtered_repr(
        fn: ta.Union[ta.Callable[[ta.Any, dc.Field, ta.Any], bool], ta.Literal['omit_none', 'omit_falsey']],
):
    def inner(cls):
        if not isinstance(cls, type) and dc.is_dataclass(cls):
            raise TypeError(cls)

        def filtered_repr(self) -> str:
            return dataclass_filtered_repr(self, fn)

        _install_dataclass_fn(cls, filtered_repr, '__repr__')

        return cls

    return inner


#


def dataclass_terse_repr(obj: ta.Any) -> str:
    return f'{obj.__class__.__qualname__}({", ".join(repr(getattr(obj, f.name)) for f in dc.fields(obj))})'


def install_dataclass_terse_repr():
    def inner(cls):
        if not isinstance(cls, type) and dc.is_dataclass(cls):
            raise TypeError(cls)

        def terse_repr(self) -> str:
            return dataclass_terse_repr(self)

        _install_dataclass_fn(cls, terse_repr, '__repr__')

        return cls

    return inner


##


def dataclass_descriptor_method(*bind_attrs: str, bind_owner: bool = False) -> ta.Callable:
    if not bind_attrs:
        def __get__(self, instance, owner=None):  # noqa
            return self

    elif bind_owner:
        def __get__(self, instance, owner=None):  # noqa
            # Guaranteed to return a new instance even with no attrs
            return dc.replace(self, **{
                a: v.__get__(instance, owner) if (v := getattr(self, a)) is not None else None
                for a in bind_attrs
            })

    else:
        def __get__(self, instance, owner=None):  # noqa
            if instance is None:
                return self

            # Guaranteed to return a new instance even with no attrs
            return dc.replace(self, **{
                a: v.__get__(instance, owner) if (v := getattr(self, a)) is not None else None
                for a in bind_attrs
            })

    return __get__


##


def install_dataclass_kw_only_init():
    def inner(cls):
        if not isinstance(cls, type) and dc.is_dataclass(cls):
            raise TypeError(cls)

        real_init = cls.__init__  # type: ignore[misc]

        flds = dc.fields(cls)  # noqa

        if any(f.name == 'self' for f in flds):
            self_name = '__dataclass_self__'
        else:
            self_name = 'self'

        src = '\n'.join([
            'def __init__(',
            f'    {self_name},',
            '    *,',
            *[
                ''.join([
                    f'    {f.name}: __dataclass_type_{f.name}__',
                    f' = __dataclass_default_{f.name}__' if f.default is not dc.MISSING else '',
                    ',',
                ])
                for f in flds
            ],
            ') -> __dataclass_None__:',
            '    __dataclass_real_init__(',
            f'        {self_name},',
            *[
                f'        {f.name}={f.name},'
                for f in flds
            ],
            '    )',
        ])

        ns: dict = {
            '__dataclass_None__': None,
            '__dataclass_real_init__': real_init,
            **{
                f'__dataclass_type_{f.name}__': f.type
                for f in flds
            },
            **{
                f'__dataclass_default_{f.name}__': f.default
                for f in flds
                if f.default is not dc.MISSING
            },
        }

        exec(src, ns)

        kw_only_init = ns['__init__']

        functools.update_wrapper(kw_only_init, real_init)

        _install_dataclass_fn(cls, kw_only_init, '__init__')

        return cls

    return inner


##


@dc.dataclass()
class DataclassFieldRequiredError(Exception):
    name: str


def dataclass_field_required(name: str) -> ta.Callable[[], ta.Any]:
    def inner() -> ta.NoReturn:
        raise DataclassFieldRequiredError(name)
    return inner



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
from ....lite.check import check
from ..compression.compressors import IoPipelineHttpObjectCompressor
from ..encoders import IoPipelineHttpObjectEncoder
from ..objects import IoPipelineHttpMessageHead
from ..responses import IoPipelineHttpResponseHead
from ..responses import IoPipelineHttpResponseObjects


##


class IoPipelineHttpResponseEncoder(IoPipelineHttpResponseObjects, IoPipelineHttpObjectEncoder):
    def _encode_head_line(self, head: IoPipelineHttpMessageHead) -> bytes:
        head = check.isinstance(head, IoPipelineHttpResponseHead)
        version_str = f'HTTP/{head.version.major}.{head.version.minor}'
        return f'{version_str} {head.status} {head.reason}\r\n'.encode('ascii')


##


class IoPipelineHttpResponseCompressor(IoPipelineHttpResponseObjects, IoPipelineHttpObjectCompressor):
    pass



"""
https://peps.python.org/pep-0249/
"""
import enum
import typing as ta


##


DbapiTypeCode: ta.TypeAlias = ta.Any | None

DbapiColumnDescription: ta.TypeAlias = tuple[
    str,  # name
    DbapiTypeCode,  # type_code
    int | None,  # display_size
    int | None,  # internal_size
    int | None,  # precision
    int | None,  # scale
    bool | None,  # null_ok
]


class DbapiColumnDescription_(ta.NamedTuple):  # noqa
    name: str
    type_code: DbapiTypeCode
    display_size: int | None
    internal_size: int | None
    precision: int | None
    scale: int | None
    null_ok: bool | None

    @classmethod
    def of(cls, obj: ta.Any) -> 'DbapiColumnDescription_':
        if isinstance(obj, cls):
            return obj
        else:
            return cls(*obj[:len(cls._fields)])


##


class DbapiCursor(ta.Protocol):
    @property
    def description(self) -> ta.Sequence[DbapiColumnDescription] | None: ...

    @property
    def rowcount(self) -> int: ...

    # optional:
    # def callproc(self, procname: str, parameters: Sequence[ta.Any] = ...) -> Sequence[ta.Any]: ...

    def close(self) -> object: ...

    def execute(
        self,
        operation: str,
        parameters: ta.Sequence[ta.Any] | ta.Mapping[str, ta.Any] = ...,
    ) -> object: ...

    def executemany(
        self,
        operation: str,
        seq_of_parameters: ta.Sequence[ta.Sequence[ta.Any]],
    ) -> object: ...

    def fetchone(self) -> ta.Sequence[ta.Any] | None: ...

    def fetchmany(self, size: int = ...) -> ta.Sequence[ta.Sequence[ta.Any]]: ...

    def fetchall(self) -> ta.Sequence[ta.Sequence[ta.Any]]: ...

    arraysize: int

    def setinputsizes(self, sizes: ta.Sequence[DbapiTypeCode | int | None]) -> object: ...

    def setoutputsize(self, size: int, column: int = ...) -> object: ...


class HasRowNumberDbapiCursor(DbapiCursor, ta.Protocol):
    # optional:
    @property
    def rownumber(self) -> int: ...


class HasLastRowIdDbapiCursor(DbapiCursor, ta.Protocol):
    # optional:
    @property
    def lastrowid(self) -> ta.Any: ...


##


class DbapiConnection(ta.Protocol):
    def close(self) -> object: ...

    def commit(self) -> object: ...

    # optional:
    def rollback(self) -> ta.Any: ...

    # optional:
    autocommit: int

    def cursor(self) -> 'DbapiCursor': ...


##


class DbapiThreadSafety(enum.IntEnum):
    NONE = 0
    MODULE = 1
    CONNECTION = 2
    CURSOR = 3


class DbapiModule(ta.Protocol):
    def connect(self, *args: ta.Any, **kwargs: ta.Any) -> DbapiConnection: ...

    #

    @property
    def apilevel(self) -> str: ...

    @property
    def threadsafety(self) -> int: ...

    @property
    def paramstyle(self) -> str: ...

    #

    @property
    def Warning(self) -> type[Exception]: ...  # noqa

    @property
    def Error(self) -> type[Exception]: ...  # noqa

    @property
    def InterfaceError(self) -> type[Exception]: ...  # noqa

    @property
    def DatabaseError(self) -> type[Exception]: ...  # noqa

    @property
    def DataError(self) -> type[Exception]: ...  # noqa

    @property
    def OperationalError(self) -> type[Exception]: ...  # noqa

    @property
    def IntegrityError(self) -> type[Exception]: ...  # noqa

    @property
    def InternalError(self) -> type[Exception]: ...  # noqa

    @property
    def ProgrammingError(self) -> type[Exception]: ...  # noqa

    @property
    def NotSupportedError(self) -> type[Exception]: ...  # noqa

    #

    def Date(self, year: ta.Any, month: ta.Any, day: ta.Any) -> ta.Any: ...  # noqa

    def Time(self, hour: ta.Any, minute: ta.Any, second: ta.Any) -> ta.Any: ...  # noqa

    def Timestamp(self, year, month, day, hour, minute, second) -> ta.Any: ...  # noqa

    def DateFromTicks(self, ticks: ta.Any) -> ta.Any: ...  # noqa

    def TimeFromTicks(self, ticks: ta.Any) -> ta.Any: ...  # noqa

    def TimestampFromTicks(self, ticks: ta.Any) -> ta.Any: ...  # noqa

    def Binary(self, string: ta.Any) -> ta.Any: ...  # noqa

    @property
    def STRING(self) -> type: ...  # noqa

    @property
    def BINARY(self) -> type: ...  # noqa

    @property
    def NUMBER(self) -> type: ...  # noqa

    @property
    def DATETIME(self) -> type: ...  # noqa

    @property
    def ROWID(self) -> type: ...  # noqa



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import collections
import dataclasses as dc
import enum
import typing as ta

from ...lite.abstract import Abstract
from ...lite.check import check
from ...lite.namespaces import NamespaceClass
from .errors import ContextInvalidatedIoPipelineError
from .errors import MessageNotPropagatedIoPipelineError
from .errors import MessageReachedTerminalIoPipelineError
from .errors import SawFinalInputIoPipelineError
from .errors import SawFinalOutputIoPipelineError
from .errors import SawInitialInputIoPipelineError
from .errors import UnhandleableIoPipelineError


F = ta.TypeVar('F')
T = ta.TypeVar('T')

IoPipelineHandlerFn = ta.Callable[['IoPipelineHandlerContext', F], T]  # ta.TypeAlias

IoPipelineHandlerT = ta.TypeVar('IoPipelineHandlerT', bound='IoPipelineHandler')
ShareableIoPipelineHandlerT = ta.TypeVar('ShareableIoPipelineHandlerT', bound='ShareableIoPipelineHandler')  # noqa

IoPipelineMetadataT = ta.TypeVar('IoPipelineMetadataT', bound='IoPipelineMetadata')


##


class IoPipelineMessages(NamespaceClass):
    """Standard messages sent through a pipeline."""

    #

    class NeverInbound(Abstract):
        pass

    class NeverOutbound(Abstract):
        pass

    #

    class MayPropagate(Abstract):
        """
        These *may* be propagated all the way through the pipeline without being an error. These will be silently
        dropped when fed inbound and reaching the innermost pipeline position, but will still be emitted as pipeline
        output when fed outbound.
        """

    class MustPropagate(MayPropagate, Abstract):
        """
        These *must* be propagated all the way through the pipeline when sent in either direction. This is enforced via
        object identity - the same *instance* of the message must be seen at the end of the pipeline to be considered
        caught. This is intentional.
        """

    #

    class Pinning(Abstract):
        @property
        @abc.abstractmethod
        def pinned(self) -> ta.Optional[ta.Sequence['IoPipelineMessages.MustPropagate']]:
            raise NotImplementedError

    #

    @ta.final
    @dc.dataclass(frozen=True, eq=False)
    class InitialInput(NeverOutbound, MustPropagate):  # ~ Netty `ChannelInboundHandler::channelActive`
        """Signals that the inbound stream has begun producing input (`connected`)."""

        def __repr__(self) -> str:
            return f'{type(self).__name__}@{id(self):x}()'

    @ta.final
    @dc.dataclass(frozen=True, eq=False)
    class FinalInput(NeverOutbound, MustPropagate):  # ~ Netty `ChannelInboundHandler::channelInactive`
        """Signals that the inbound stream has produced its final message (`eof`)."""

        def __repr__(self) -> str:
            return f'{type(self).__name__}@{id(self):x}()'

    @ta.final
    @dc.dataclass(frozen=True, eq=False)
    class FinalOutput(NeverInbound, MustPropagate):  # ~ Netty `ChannelOutboundHandler::close`
        """Signals that the outbound stream has produced its final message (`close`)."""

        def __repr__(self) -> str:
            return f'{type(self).__name__}@{id(self):x}()'

    #

    @ta.final
    @dc.dataclass(frozen=True)
    class Error(NeverOutbound):
        """Signals an exception occurred in the pipeline."""

        exc: BaseException

        direction: ta.Optional['IoPipelineDirection'] = None
        handler: ta.Optional['IoPipelineHandlerRef'] = None

    #

    class Completable(Abstract, ta.Generic[T]):
        # Management of completable state is implemented as 'hidden' / dynamic attributes to allow mixing in with
        # otherwise frozen dataclasses.

        # _completion_state: ta.Literal['pending', 'succeeded', 'failed'] = 'pending'
        # _completion_: _Completion

        @ta.final
        class _Completion:
            result: ta.Any
            exc: ta.Optional[BaseException]
            listeners: ta.Optional[ta.List[ta.Callable[[ta.Any], None]]] = None

        def is_done(self) -> bool:
            try:
                cps = self._completion_state  # type: ignore[attr-defined]
            except AttributeError:
                return False
            return cps != 'pending'

        def is_succeeded(self) -> bool:
            try:
                cps = self._completion_state  # type: ignore[attr-defined]
            except AttributeError:
                return False
            return cps == 'succeeded'

        def get_result(self) -> T:
            check.state(self._completion_state == 'succeeded')  # type: ignore[attr-defined]

            return self._completion_.result  # type: ignore[attr-defined]

        def is_failed(self) -> bool:
            try:
                cps = self._completion_state  # type: ignore[attr-defined]
            except AttributeError:
                return False
            return cps == 'failed'

        def get_exception(self) -> ta.Optional[BaseException]:
            check.state(self._completion_state == 'failed')  # type: ignore[attr-defined]

            return self._completion_.exc  # type: ignore[attr-defined]

        def _completion(self) -> _Completion:
            try:
                return self._completion_  # type: ignore[attr-defined]
            except AttributeError:
                pass

            cpl = IoPipelineMessages.Completable._Completion()  # noqa
            object.__setattr__(self, '_completion_', cpl)
            return cpl

        def add_listener(self, fn: ta.Callable[['IoPipelineMessages.Completable[T]'], None]) -> None:
            check.state(not self.is_done())

            cpl = self._completion()
            if (lst := cpl.listeners) is None:
                lst = cpl.listeners = []
            lst.append(fn)

        def set_succeeded(self, result: T) -> None:
            check.state(not self.is_done())

            object.__setattr__(self, '_completion_state', 'succeeded')

            try:
                cpl = self._completion_  # type: ignore[attr-defined]
            except AttributeError:
                return

            cpl.result = result
            if (lst := cpl.listeners) is not None:
                for fn in lst:
                    fn(self)

            object.__delattr__(self, '_completion_')

        def set_failed(self, exc: ta.Optional[BaseException] = None) -> None:
            check.state(not self.is_done())

            object.__setattr__(self, '_completion_state', 'failed')

            try:
                cpl = self._completion_  # type: ignore[attr-defined]
            except AttributeError:
                return

            cpl.exc = exc
            if (lst := cpl.listeners) is not None:
                for fn in lst:
                    fn(self)

            object.__delattr__(self, '_completion_')

    #

    @ta.final
    @dc.dataclass(frozen=True)
    class Defer(NeverInbound, Pinning, Completable[T], ta.Generic[T]):
        fn: ta.Union[
            ta.Callable[['IoPipelineHandlerContext'], T],
            ta.Callable[[], T],
        ]

        no_context: bool = False

        def __repr__(self) -> str:
            return f'{type(self).__name__}@{id(self):x}({self.fn!r}{", no_context=True" if self.no_context else ""})'

        # _: dc.KW_ONLY

        _ctx: ta.Optional['IoPipelineHandlerContext'] = dc.field(default=None, repr=False)

        _pinned: ta.Optional[ta.Sequence['IoPipelineMessages.MustPropagate']] = dc.field(default=None, repr=False)

        @property
        def pinned(self) -> ta.Optional[ta.Sequence['IoPipelineMessages.MustPropagate']]:
            return self._pinned


##


class IoPipelineHandlerNotification(Abstract):  # ~ Netty `ChannelHandler` methods
    """
    Directionless, private events sent to a specific handler that are not to be forwarded to any other handler in either
    direction.
    """


class IoPipelineHandlerNotifications(NamespaceClass):
    @ta.final
    @dc.dataclass(frozen=True)
    class Added(IoPipelineHandlerNotification):
        ctx: 'IoPipelineHandlerContext'

    @ta.final
    @dc.dataclass(frozen=True)
    class Removed(IoPipelineHandlerNotification):
        ctx: 'IoPipelineHandlerContext'


##


@ta.final
class IoPipelineHandlerRef(ta.Generic[T]):
    """
    Encapsulates a reference to a unique position of a handler instance in a pipeline, used at public api boundaries.

    Should the handler be removed from the relevant position in the pipeline, the ref instance becomes permanently
    invalidated.

    Note that this is definitionally identity hash/eq: given some valid ref, removing that ref from the pipeline and
    re-adding the same handler instance to the same effective position in a pipeline results in a different ref.
    """

    def __init__(self, *, _context: 'IoPipelineHandlerContext') -> None:
        self._context = _context

    @property
    def pipeline(self) -> 'IoPipeline':
        return self._context._pipeline  # noqa

    @property
    def handler(self) -> T:
        return self._context._handler  # type: ignore[return-value]  # noqa

    @property
    def name(self) -> ta.Optional[str]:
        return self._context._name  # noqa

    @property
    def invalidated(self) -> bool:
        return self._context._invalidated  # noqa

    def __repr__(self) -> str:
        return (
            f'{type(self).__name__}'
            f'{"!INVALIDATED" if self.invalidated else ""}'
            f'{f"<{self.name!r}>" if self.name is not None else ""}'
            f'<context@{id(self._context):x}>'
            f'({self.handler!r})'  # {f"@{id(self.handler):x}"})'
        )


IoPipelineHandlerRef_ = IoPipelineHandlerRef['IoPipelineHandler']  # ta.TypeAlias  # omlish-amalg-typing-no-move  # noqa


##


@ta.final
class IoPipelineHandlerContext:
    """
    The embodiment of an instance of a handler at a position in a pipeline. Passed to IoPipelineHandler methods,
    providing handler-specific access to the pipeline. As instances of `ShareableIoPipelineHandler` may validly be
    simultaneously present at multiple positions in a pipeline, a single handler may have multiple active context
    instances associated with it in any given pipeline.

    Instances of this class are considered private to a handler instance and are not to be cached or shared in any way.
    The method names reflect this: they are operations available to the handler in the context of a pipeline processing
    operation.
    """

    def __init__(
            self,
            *,
            _pipeline: 'IoPipeline',
            _handler: 'IoPipelineHandler',

            _name: ta.Optional[str] = None,
    ) -> None:
        super().__init__()

        self._pipeline: ta.Final[IoPipeline] = _pipeline
        self._handler: ta.Final[IoPipelineHandler] = _handler

        self._name: ta.Final[ta.Optional[str]] = _name

        self._ref: IoPipelineHandlerRef_ = IoPipelineHandlerRef(_context=self)

        hty = type(_handler)
        self._handles_inbound = hty.inbound is not IoPipelineHandler.inbound
        self._handles_outbound = hty.outbound is not IoPipelineHandler.outbound

    _next_in: 'IoPipelineHandlerContext'  # 'next'
    _next_out: 'IoPipelineHandlerContext'  # 'prev'

    def __repr__(self) -> str:
        return (
            f'{type(self).__name__}@{id(self):x}'
            f'{"!INVALIDATED" if self._invalidated else ""}'
            f'{f"<{self._name!r}>" if self._name is not None else ""}'
            f'<pipeline@{id(self.pipeline):x}>'
            f'({self._handler!r})'  # @{id(self._handler):x})'
        )

    @property
    def ref(self) -> IoPipelineHandlerRef_:
        return self._ref

    @property
    def pipeline(self) -> 'IoPipeline':
        return self._pipeline

    @property
    def services(self) -> 'IoPipelineServices':  # noqa
        return self._pipeline._services  # noqa

    @property
    def handler(self) -> 'IoPipelineHandler':
        return self._handler

    @property
    def name(self) -> ta.Optional[str]:
        return self._name

    #

    def defer(
            self,
            fn: ta.Callable[['IoPipelineHandlerContext'], T],
            *,
            pin: ta.Optional[ta.Sequence[IoPipelineMessages.MustPropagate]] = None,
    ) -> IoPipelineMessages.Defer[T]:
        return self._pipeline._defer(self, fn, pin=pin)  # noqa

    def defer_no_context(
            self,
            fn: ta.Callable[[], T],
            *,
            pin: ta.Optional[ta.Sequence[IoPipelineMessages.MustPropagate]] = None,
    ) -> IoPipelineMessages.Defer[T]:
        return self._pipeline._defer(self, fn, no_context=True, pin=pin)  # noqa

    #

    _invalidated = False

    @property
    def invalidated(self) -> bool:
        return self._invalidated

    #

    @ta.final
    @dc.dataclass(frozen=True)
    class StorageKey(ta.Generic[T]):
        name: str

    @ta.final
    class Storage:
        def __init__(self) -> None:
            self.__dict: ta.Dict[IoPipelineHandlerContext.StorageKey, ta.Any] = {}

        @property
        def dict(self) -> ta.Dict['IoPipelineHandlerContext.StorageKey', ta.Any]:
            return self.__dict

        def __getitem__(self, key: 'IoPipelineHandlerContext.StorageKey[T]') -> T:
            return self.__dict[key]

        @ta.overload
        def get(
                self,
                key: 'IoPipelineHandlerContext.StorageKey[T]',
                default: T,
                /,
        ) -> T:
            ...

        @ta.overload
        def get(
                self,
                key: 'IoPipelineHandlerContext.StorageKey[T]',
                default: ta.Optional[T] = None,
                /,
        ) -> ta.Optional[T]:
            ...

        def get(self, key, default=None, /):
            return self.__dict.get(key, default)

        def __setitem__(self, key: 'IoPipelineHandlerContext.StorageKey[T]', value: T) -> None:
            self.__dict[key] = value

        def __delitem__(self, key: 'IoPipelineHandlerContext.StorageKey[T]') -> None:
            del self.__dict[key]

        def __len__(self) -> int:
            return len(self.__dict)

        def __contains__(self, key: 'IoPipelineHandlerContext.StorageKey[T]') -> bool:
            return key in self.__dict

        def __iter__(self) -> ta.Iterator['IoPipelineHandlerContext.StorageKey[T]']:
            return iter(self.__dict)

        def items(self) -> ta.Iterator[ta.Tuple['IoPipelineHandlerContext.StorageKey[T]', T]]:
            return iter(self.__dict.items())

    _storage_: Storage

    @property
    def storage(self) -> Storage:
        try:
            return self._storage_
        except AttributeError:
            pass
        self._storage_ = ret = IoPipelineHandlerContext.Storage()
        return ret

    #

    def _notify(self, no: IoPipelineHandlerNotification) -> None:
        check.isinstance(no, IoPipelineHandlerNotification)
        check.state(self._pipeline._execution_depth > 0)  # noqa

        self._handler.notify(self, no)

    ##
    # Feeding `type`'s is forbidden as it's almost always going to be an error - usually forgetting to instantiate a
    # marker dataclass)

    _FORBIDDEN_INBOUND_TYPES: ta.ClassVar[ta.Tuple[type, ...]] = (
        IoPipelineMessages.NeverInbound,
        IoPipelineHandlerNotification,
        type,
        type(None),
    )

    def _inbound(self, msg: ta.Any) -> None:
        check.state(not self._invalidated, ContextInvalidatedIoPipelineError)
        check.state(self._pipeline._state == IoPipeline.State.READY and self._pipeline._execution_depth > 0)  # noqa

        check.not_isinstance(msg, self._FORBIDDEN_INBOUND_TYPES)

        if isinstance(msg, IoPipelineMessages.MustPropagate):
            self._pipeline._propagation.add_must(self, 'inbound', msg)  # noqa

        try:
            self._handler.inbound(self, msg)

        except self._pipeline._all_never_handle_exceptions:  # type: ignore[misc]  # noqa
            raise

        except BaseException as e:
            if self._handling_error or self._pipeline._config.raise_immediately:  # noqa
                raise
            self._handle_error(e, 'inbound')

    _FORBIDDEN_OUTBOUND_TYPES: ta.ClassVar[ta.Tuple[type, ...]] = (
        IoPipelineMessages.NeverOutbound,
        IoPipelineHandlerNotification,
        type,
        type(None),
    )

    def _outbound(self, msg: ta.Any) -> None:
        check.state(not self._invalidated, ContextInvalidatedIoPipelineError)
        check.state(self._pipeline._state == IoPipeline.State.READY and self._pipeline._execution_depth > 0)  # noqa

        check.not_isinstance(msg, self._FORBIDDEN_OUTBOUND_TYPES)

        if isinstance(msg, IoPipelineMessages.MustPropagate):
            self._pipeline._propagation.add_must(self, 'outbound', msg)  # noqa

        try:
            self._handler.outbound(self, msg)

        except self._pipeline._all_never_handle_exceptions:  # type: ignore[misc]  # noqa
            raise

        except BaseException as e:
            if self._handling_error or self._pipeline._config.raise_immediately:  # noqa
                raise
            self._handle_error(e, 'outbound')

    #

    def _run_deferred(self, dfl: IoPipelineMessages.Defer) -> None:
        check.state(not self._invalidated, ContextInvalidatedIoPipelineError)
        check.state(self._pipeline._state == IoPipeline.State.READY and self._pipeline._execution_depth > 0)  # noqa

        check.state(dfl._ctx is self)  # noqa

        try:
            if dfl.no_context:
                res = dfl.fn()  # type: ignore[call-arg]
            else:
                res = dfl.fn(self)  # type: ignore[call-arg]

        except self._pipeline._all_never_handle_exceptions:  # type: ignore[misc]  # noqa
            raise

        except BaseException as e:  # noqa
            dfl.set_failed(e)

            if self._handling_error or self._pipeline._config.raise_immediately:  # noqa
                raise
            self._handle_error(e, 'inbound')

        else:
            dfl.set_succeeded(res)

    #

    _handling_error: bool = False

    def _handle_error(self, e: BaseException, direction: 'IoPipelineDirection') -> None:
        check.state(not self._handling_error)
        self._handling_error = True

        try:
            try:
                self.feed_in(IoPipelineMessages.Error(e, direction, self._ref))

            except self._pipeline._all_never_handle_exceptions:  # type: ignore[misc]  # noqa
                raise

            except BaseException as e2:  # noqa
                raise

        finally:
            self._handling_error = False

    ##
    # The following overloads attempts to catch invalid inputs statically, but there's no explicit way to do this in
    # mypy - the following trick only works if there's an unconditional statement after the attempted calls, but it's
    # better than nothing.

    @ta.overload
    def feed_in(
            self,
            msg: ta.Union[
                IoPipelineMessages.NeverInbound,
                IoPipelineHandlerNotification,
                type,
                None,
            ],
    ) -> 'ta.Never':
        ...

    @ta.overload
    def feed_in(self, msg: object) -> None:
        ...

    def feed_in(self, msg):  # ~ Netty `ChannelInboundInvoker::fireChannelRead`
        nxt = self._next_in
        while not nxt._handles_inbound:  # noqa
            nxt = nxt._next_in  # noqa
        nxt._inbound(msg)  # noqa

    @ta.overload
    def feed_out(
            self,
            msg: ta.Union[
                IoPipelineMessages.NeverOutbound,
                IoPipelineHandlerNotification,
                type,
                None,
            ],
    ) -> 'ta.Never':
        ...

    @ta.overload
    def feed_out(self, msg: object) -> None:
        ...

    def feed_out(self, msg):  # ~ Netty `ChannelOutboundInvoker::write`
        nxt = self._next_out  # noqa
        while not nxt._handles_outbound:  # noqa
            nxt = nxt._next_out  # noqa
        nxt._outbound(msg)  # noqa

    #

    def feed_final_output(self) -> None:
        self.feed_out(IoPipelineMessages.FinalOutput())

    #

    def mark_propagated(
            self,
            direction: 'IoPipelineDirection',
            msg: IoPipelineMessages.MustPropagate,
    ) -> None:
        self._pipeline._propagation.remove_must(self, direction, msg)  # noqa


##


class IoPipelineHandler(Abstract):
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        if not (
            cls.__hash__ is object.__hash__ and
            cls.__eq__ is object.__eq__ and
            cls.__ne__ is object.__ne__
        ):
            raise TypeError(
                f'IoPipelineHandler subclass {cls.__name__} must not override __hash__, __eq__ or __ne__',
            )

    #

    def notify(self, ctx: IoPipelineHandlerContext, no: IoPipelineHandlerNotification) -> None:
        pass

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        ctx.feed_in(msg)

    def outbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        ctx.feed_out(msg)


class ShareableIoPipelineHandler(IoPipelineHandler, Abstract):
    pass


##


IoPipelineDirection = ta.Literal[  # ta.TypeAlias  # omlish-amalg-typing-no-move
    'inbound',
    'outbound',
]

IoPipelineDirectionOrDuplex = ta.Literal[  # ta.TypeAlias  # omlish-amalg-typing-no-move
    IoPipelineDirection,
    'duplex',
]

IoPipelineUpdate = ta.Literal[  # ta.TypeAlias  # omlish-amalg-typing-no-move
    'added',
    'removed',
]

IoPipelineHandlerUpdate = ta.Literal[  # ta.TypeAlias  # omlish-amalg-typing-no-move
    'adding',
    'added',
    'removing',
    'removed',
]


##


class IoPipelineService(Abstract):
    def pipeline_update(self, pipeline: 'IoPipeline', kind: IoPipelineUpdate) -> None:
        pass

    def handler_update(self, handler_ref: IoPipelineHandlerRef, kind: IoPipelineHandlerUpdate) -> None:
        pass

    def pipeline_enter(self, pipeline: 'IoPipeline') -> None:
        pass

    def pipeline_exit(self, pipeline: 'IoPipeline') -> None:
        pass


@ta.final
class IoPipelineServices:
    def __init__(self, lst: ta.Sequence[IoPipelineService]) -> None:
        self._lst = lst

        self._by_type_cache: ta.Dict[type, ta.Sequence[ta.Any]] = {}
        self._single_by_type_cache: ta.Dict[type, ta.Optional[ta.Any]] = {}

        self._handles_pipeline_update = handles_pipeline_update = []
        self._handles_handler_update = handles_handler_update = []
        self._handles_pipeline_enter = handles_pipeline_enter = []
        self._handles_pipeline_exit = handles_pipeline_exit = []

        for svc in lst:
            sty = type(svc)
            if sty.pipeline_update is not IoPipelineService.pipeline_update:
                handles_pipeline_update.append(sty)
            if sty.handler_update is not IoPipelineService.handler_update:
                handles_handler_update.append(sty)
            if sty.pipeline_enter is not IoPipelineService.pipeline_enter:
                handles_pipeline_enter.append(sty)
            if sty.pipeline_exit is not IoPipelineService.pipeline_exit:
                handles_pipeline_exit.append(sty)

    _handles_pipeline_update: ta.Sequence[IoPipelineService]
    _handles_handler_update: ta.Sequence[IoPipelineService]
    _handles_pipeline_enter: ta.Sequence[IoPipelineService]
    _handles_pipeline_exit: ta.Sequence[IoPipelineService]

    _EMPTY: ta.ClassVar['IoPipelineServices']

    @classmethod
    def of(cls, obj: ta.Union['IoPipelineServices', ta.Sequence[IoPipelineService]]) -> 'IoPipelineServices':  # noqa
        if isinstance(obj, cls):
            return obj
        elif not obj:
            return cls._EMPTY
        else:
            return cls(list(obj))

    def __len__(self) -> int:
        return len(self._lst)

    def __iter__(self) -> ta.Iterator[IoPipelineService]:
        return iter(self._lst)

    def __contains__(self, item: IoPipelineService) -> bool:
        return item in self._lst

    @dc.dataclass(frozen=True)
    class ServiceType(ta.Generic[T]):
        """This is entirely just a workaround for mypy's `type-abstract` deficiency."""

        ty: ta.Type[T]

    def find_all(self, ty: ta.Union[ServiceType[T], ta.Type[T]]) -> ta.Sequence[T]:
        if isinstance(ty, self.ServiceType):
            ty = ty.ty

        try:
            return self._by_type_cache[ty]
        except KeyError:
            pass

        self._by_type_cache[ty] = ret = [svc for svc in self._lst if isinstance(svc, ty)]
        return ret

    def find(self, ty: ta.Union[ServiceType[T], ta.Type[T]]) -> ta.Optional[T]:
        if isinstance(ty, self.ServiceType):
            ty = ty.ty

        try:
            return self._single_by_type_cache[ty]
        except KeyError:
            pass

        self._single_by_type_cache[ty] = ret = check.opt_single(self.find_all(ty))
        return ret

    def __getitem__(self, ty: ta.Union[ServiceType[T], ta.Type[T]]) -> T:
        if (svc := self.find(ty)) is None:
            raise KeyError(ty)
        return svc


IoPipelineServices._EMPTY = IoPipelineServices([])  # noqa


##


class IoPipelineMetadata(Abstract):
    pass


@ta.final
class IoPipelineMetadatas:
    def __init__(self, lst: ta.Sequence[IoPipelineMetadata]) -> None:
        dct: ta.Dict[type, ta.Any] = {}
        for md in lst:
            mty = type(md)
            check.not_in(mty, dct)
            dct[mty] = md
        self._dct = dct

    _EMPTY: ta.ClassVar['IoPipelineMetadatas']

    @classmethod
    def of(cls, obj: ta.Union['IoPipelineMetadatas', ta.Sequence[IoPipelineMetadata]]) -> 'IoPipelineMetadatas':  # noqa
        if isinstance(obj, cls):
            return obj
        elif not obj:
            return cls._EMPTY
        else:
            return cls(list(obj))

    def __len__(self) -> int:
        return len(self._dct)

    def __contains__(self, ty: ta.Type[IoPipelineMetadata]) -> bool:
        return ty in self._dct

    def __iter__(self) -> ta.Iterator[IoPipelineMetadata]:
        return iter(self._dct.values())

    @dc.dataclass(frozen=True)
    class MetadataType(ta.Generic[IoPipelineMetadataT]):
        """This is entirely just a workaround for mypy's `type-abstract` deficiency."""

        ty: ta.Type[IoPipelineMetadataT]

    def __getitem__(
            self,
            ty: ta.Union[
                MetadataType[IoPipelineMetadataT],
                ta.Type[IoPipelineMetadataT],
            ],
    ) -> IoPipelineMetadataT:
        if isinstance(ty, self.MetadataType):
            ty = ty.ty

        return self._dct[ty]

    @ta.overload
    def get(
            self,
            ty: ta.Union[
                MetadataType[IoPipelineMetadataT],
                ta.Type[IoPipelineMetadataT],
            ],
            default: IoPipelineMetadataT,
            /,
    ) -> IoPipelineMetadataT:
        ...

    @ta.overload
    def get(
            self,
            ty: ta.Union[
                MetadataType[IoPipelineMetadataT],
                ta.Type[IoPipelineMetadataT],
            ],
            default: ta.Optional[IoPipelineMetadataT] = None,
            /,
    ) -> ta.Optional[IoPipelineMetadataT]:
        ...

    def get(self, ty, default=None, /):
        if isinstance(ty, self.MetadataType):
            ty = ty.ty

        return self._dct.get(ty, default)


IoPipelineMetadatas._EMPTY = IoPipelineMetadatas([])  # noqa


##


@ta.final
class _IoPipelinePropagation:
    @dc.dataclass()
    class _PendingMustEntry:
        msg: ta.Any
        direction: IoPipelineDirection
        last_seen: IoPipelineHandlerContext
        pinned_by: ta.Optional[IoPipelineMessages.Pinning] = None

    def __init__(self, p: 'IoPipeline') -> None:
        self._p = p

        if not self._p._config.disable_propagation_checking:  # noqa
            self._pending_must: ta.Final[ta.Dict[int, _IoPipelinePropagation._PendingMustEntry]] = {}

    def add_must(
            self,
            ctx: IoPipelineHandlerContext,
            direction: IoPipelineDirection,
            msg: IoPipelineMessages.MustPropagate,
    ) -> None:
        if self._p._config.disable_propagation_checking:  # noqa
            return

        i = id(msg)
        try:
            x = self._pending_must[i]
        except KeyError:
            self._pending_must[i] = _IoPipelinePropagation._PendingMustEntry(  # noqa
                msg,
                direction,
                ctx,
            )
            return

        check.is_(msg, x.msg)
        check.equal(direction, x.direction)
        check.state(x.pinned_by is None)
        x.last_seen = ctx

    def pin_musts(
            self,
            pinning: IoPipelineMessages.Pinning,
    ) -> None:
        if self._p._config.disable_propagation_checking or not (lst := pinning.pinned):  # noqa
            return

        for msg in lst:
            x = self._pending_must[id(msg)]
            check.none(x.pinned_by)
            x.pinned_by = pinning

    def unpin_musts(
            self,
            pinning: IoPipelineMessages.Pinning,
    ) -> None:
        if self._p._config.disable_propagation_checking or not (lst := pinning.pinned):  # noqa
            return

        for msg in lst:
            x = self._pending_must[id(msg)]
            check.is_(x.pinned_by, pinning)
            x.pinned_by = None

    def remove_must(
            self,
            ctx: IoPipelineHandlerContext,
            direction: IoPipelineDirection,
            msg: IoPipelineMessages.MustPropagate,
    ) -> None:
        if self._p._config.disable_propagation_checking:  # noqa
            return

        i = id(msg)
        try:
            x = self._pending_must.pop(i)
        except KeyError:
            raise MessageNotPropagatedIoPipelineError.new_single(
                direction,
                msg,
                last_seen=ctx._ref,  # noqa
            ) from None

        if (
                x.msg is not msg or
                x.direction != direction or
                x.pinned_by is not None
        ):
            raise MessageNotPropagatedIoPipelineError.new_single(
                direction,
                msg,
                last_seen=ctx._ref,  # noqa
            )

    def check_and_clear(self) -> None:
        if self._p._config.disable_propagation_checking:  # noqa
            return

        if not self._pending_must:
            return

        il: ta.List[ta.Tuple[ta.Any, ta.Any]] = []
        ol: ta.List[ta.Tuple[ta.Any, ta.Any]] = []

        for x in self._pending_must.values():
            if x.pinned_by is None:
                (il if x.direction == 'inbound' else ol).append((x.msg, x.last_seen._ref))  # noqa

        if not (il or ol):
            return

        e = MessageNotPropagatedIoPipelineError.new(
            inbound_with_last_seen=il,
            outbound_with_last_seen=ol,
        )

        for lst in (il, ol):
            for msg, _ in lst:
                del self._pending_must[id(msg)]

        raise e


##


@ta.final
class IoPipeline:
    @ta.final
    @dc.dataclass(frozen=True)
    class Config:
        # TODO: 'close'? 'deadletter'? combination? composition? ...
        inbound_terminal: ta.Literal['drop', 'raise'] = 'raise'

        disable_propagation_checking: bool = False

        raise_immediately: bool = False

        def __post_init__(self) -> None:
            check.in_(self.inbound_terminal, ('drop', 'raise'))

        #

        def update(self, **kwargs: ta.Any) -> 'IoPipeline.Config':
            return dc.replace(self, **kwargs)

        DEFAULT: ta.ClassVar['IoPipeline.Config']

    Config.DEFAULT = Config()

    #

    @ta.final
    @dc.dataclass(frozen=True)
    class Spec:
        # Initial handlers are optional - handlers may be freely added and removed later.
        handlers: ta.Sequence[IoPipelineHandler] = ()

        config: 'IoPipeline.Config' = dc.field(default_factory=lambda: IoPipeline.Config.DEFAULT)

        # _: dc.KW_ONLY

        # Metadata and ervices are fixed for the lifetime of the pipeline.
        metadata: ta.Union[ta.Sequence[IoPipelineMetadata], IoPipelineMetadatas] = ()
        services: ta.Union[ta.Sequence[IoPipelineService], IoPipelineServices] = ()

        #

        def update_config(self, **kwargs: ta.Any) -> 'IoPipeline.Spec':
            return dc.replace(self, config=self.config.update(**kwargs))

    @classmethod
    def new(
            cls,
            handlers: ta.Sequence[IoPipelineHandler] = (),
            config: 'IoPipeline.Config' = Config.DEFAULT,
            *,
            metadata: ta.Union[ta.Sequence[IoPipelineMetadata], IoPipelineMetadatas] = (),
            services: ta.Union[ta.Sequence[IoPipelineService], IoPipelineServices] = (),
    ) -> 'IoPipeline':
        return cls(IoPipeline.Spec(
            handlers=handlers,
            config=config,
            metadata=metadata,
            services=services,
        ))

    #

    def __init__(
            self,
            spec: Spec,
            *,
            never_handle_exceptions: ta.Tuple[type, ...] = (),
    ) -> None:
        super().__init__()

        self._config: ta.Final[IoPipeline.Config] = spec.config
        self._never_handle_exceptions = never_handle_exceptions

        self._metadata: ta.Final[IoPipelineMetadatas] = IoPipelineMetadatas.of(spec.metadata)
        self._services: ta.Final[IoPipelineServices] = IoPipelineServices.of(spec.services)

        #

        self._all_never_handle_exceptions: ta.Tuple[type, ...] = (
            UnhandleableIoPipelineError,
            *never_handle_exceptions,
        )

        self._propagation: _IoPipelinePropagation = _IoPipelinePropagation(self)

        #

        self._output: ta.Final[IoPipeline._Output] = IoPipeline._Output()

        #

        self._outermost = outermost = IoPipelineHandlerContext(
            _pipeline=self,
            _handler=IoPipeline._Outermost(),
        )
        self._innermost = innermost = IoPipelineHandlerContext(
            _pipeline=self,
            _handler=IoPipeline._Innermost(),
        )

        # Explicitly does not form a ring, iteration past the outermost/innermost is always an error and will
        # intentionally raise AttributeError if not caught earlier.
        outermost._next_in = innermost  # noqa
        innermost._next_out = outermost  # noqa

        self._unique_contexts: ta.Final[ta.Dict[IoPipelineHandler, IoPipelineHandlerContext]] = {}
        self._shareable_contexts: ta.Final[ta.Dict[ShareableIoPipelineHandler, ta.Set[IoPipelineHandlerContext]]] = {}  # noqa

        self._contexts_by_name: ta.Final[ta.Dict[str, IoPipelineHandlerContext]] = {}

        #

        self._state = IoPipeline.State.READY

        #

        try:
            for svc in self._services._handles_pipeline_update:  # noqa
                svc.pipeline_update(self, 'added')

            for h in spec.handlers:
                self.add_innermost(h)

        except self._all_never_handle_exceptions:  # type: ignore[misc]
            raise

        except BaseException:  # noqa
            self.destroy()
            raise

    def __repr__(self) -> str:
        return f'{type(self).__name__}@{id(self):x}'

    @property
    def config(self) -> Config:
        return self._config

    ##
    # state

    class State(enum.Enum):
        NEW = 'new'
        READY = 'ready'
        DESTROYING = 'destroying'
        DESTROYED = 'destroyed'

    _state: State = State.NEW

    @property
    def state(self) -> State:
        return self._state

    @property
    def is_ready(self) -> bool:
        return self._state is IoPipeline.State.READY

    #

    _saw_any_input = False
    _saw_initial_input = False
    _saw_final_input = False
    _saw_final_output = False

    @property
    def saw_any_input(self) -> bool:
        return self._saw_any_input

    @property
    def saw_initial_input(self) -> bool:
        return self._saw_initial_input

    @property
    def saw_final_input(self) -> bool:
        return self._saw_final_input  # Note: only 'pipeline-level'

    @property
    def saw_final_output(self) -> bool:
        return self._saw_final_output

    ##
    # sub-collections

    @property
    def metadata(self) -> IoPipelineMetadatas:
        return self._metadata

    @property
    def services(self) -> IoPipelineServices:
        return self._services

    ##
    # execution

    _execution_depth = 0

    def _step_in(self) -> None:
        self._execution_depth += 1

        if self._execution_depth == 1:
            for svc in self._services._handles_pipeline_enter:  # noqa
                svc.pipeline_enter(self)

    def _step_out(self) -> None:
        check.state(self._execution_depth > 0)

        self._execution_depth -= 1

        if not self._execution_depth:
            for svc in self._services._handles_pipeline_exit:  # noqa
                svc.pipeline_exit(self)

            self._propagation.check_and_clear()

    @ta.final
    class _EnterContextManager:
        def __init__(self, p: 'IoPipeline') -> None:
            self._p = p

        def __enter__(self) -> None:
            self._p._step_in()  # noqa

        def __exit__(self, exc_type, exc_val, exc_tb) -> None:
            self._p._step_out()  # noqa

    def enter(self) -> ta.ContextManager[None]:
        return self._EnterContextManager(self)

    #

    def _notify(self, ctx: IoPipelineHandlerContext, no: IoPipelineHandlerNotification) -> None:
        self._step_in()
        try:
            ctx._notify(no)  # noqa

        finally:
            self._step_out()

    def notify(self, handler_ref: IoPipelineHandlerRef, no: IoPipelineHandlerNotification) -> None:
        ctx = handler_ref._context  # noqa
        check.is_(ctx._pipeline, self)  # noqa
        self._notify(ctx, no)

    #

    def _feed_in_to(self, ctx: IoPipelineHandlerContext, msgs: ta.Iterable[ta.Any]) -> None:
        self._step_in()
        try:
            for msg in msgs:
                if self._saw_final_input:
                    raise SawFinalInputIoPipelineError
                elif isinstance(msg, IoPipelineMessages.FinalInput):
                    self._saw_final_input = True

                if isinstance(msg, IoPipelineMessages.InitialInput):
                    if self._saw_any_input:
                        raise SawInitialInputIoPipelineError
                    check.state(not self._saw_initial_input)
                    self._saw_initial_input = True
                self._saw_any_input = True

                ctx._inbound(msg)  # noqa

        finally:
            self._step_out()

    def feed_in_to(self, handler_ref: IoPipelineHandlerRef, *msgs: ta.Any) -> None:
        # TODO: remove? internal only? used by replace-self pattern
        ctx = handler_ref._context  # noqa
        check.is_(ctx._pipeline, self)  # noqa
        self._feed_in_to(ctx, msgs)

    def feed_in(self, *msgs: ta.Any) -> None:
        self._feed_in_to(self._outermost, msgs)  # noqa

    def feed_initial_input(self) -> None:
        self._feed_in_to(self._outermost, (IoPipelineMessages.InitialInput(),))  # noqa

    def feed_final_input(self) -> None:
        self._feed_in_to(self._outermost, (IoPipelineMessages.FinalInput(),))  # noqa

    #

    def _defer(
            self,
            ctx: IoPipelineHandlerContext,
            fn: ta.Union[
                ta.Callable[[IoPipelineHandlerContext], T],
                ta.Callable[[], T],
            ],
            *,
            no_context: bool = False,
            pin: ta.Optional[ta.Sequence[IoPipelineMessages.MustPropagate]] = None,
    ) -> IoPipelineMessages.Defer[T]:
        check.is_(ctx._pipeline, self)  # noqa
        check.state(not ctx._invalidated)  # noqa

        dfl = IoPipelineMessages.Defer(
            fn,
            no_context,
            _ctx=ctx,
            _pinned=pin,
        )

        if pin:
            self._propagation.pin_musts(dfl)

        ctx.feed_out(dfl)

        return dfl

    def run_deferred(self, dfl: IoPipelineMessages.Defer) -> None:
        ctx = check.not_none(dfl._ctx)  # noqa
        check.is_(ctx._pipeline, self)  # noqa
        check.state(not ctx._invalidated)  # noqa

        self._step_in()
        try:
            if dfl._pinned:  # noqa
                self._propagation.unpin_musts(dfl)

            ctx._run_deferred(dfl)  # noqa

        finally:
            self._step_out()

    ##
    # output

    @ta.final
    class _Output:
        def __init__(self) -> None:
            self._q: ta.Final[collections.deque[ta.Any]] = collections.deque()

        def __repr__(self) -> str:
            return (
                f'{type(self).__qualname__}@{id(self):x}'
                f'<len={len(self._q)}>'
                '()'
            )

        def peek(self) -> ta.Optional[ta.Any]:
            if not self._q:
                return None

            return self._q[0]

        def poll(self) -> ta.Optional[ta.Any]:
            if not self._q:
                return None

            return self._q.popleft()

        def drain(self) -> ta.List[ta.Any]:
            out: ta.List[ta.Any] = []

            while self._q:
                out.append(self._q.popleft())

            return out

    @property
    def output(self) -> _Output:
        return self._output

    ##
    # handlers

    def _handler_update(self, ctx: IoPipelineHandlerContext, kind: IoPipelineHandlerUpdate) -> None:
        for svc in self._services._handles_handler_update:  # noqa
            svc.handler_update(ctx._ref, kind)  # noqa

    #

    def _terminal_inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:  # noqa
        if (tm := self._config.inbound_terminal) == 'drop':
            pass

        elif tm == 'raise':
            if not isinstance(msg, IoPipelineMessages.MayPropagate):
                raise MessageReachedTerminalIoPipelineError.new_single('inbound', msg)

        else:
            raise RuntimeError(f'unknown inbound terminal mode {tm}')

    def _terminal_outbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:  # noqa
        if isinstance(msg, IoPipelineMessages.FinalOutput):
            self._saw_final_output = True
        elif self._saw_final_output:
            raise SawFinalOutputIoPipelineError

        self._output._q.append(msg)  # noqa

    #

    _outermost: ta.Final[IoPipelineHandlerContext]
    _innermost: ta.Final[IoPipelineHandlerContext]

    def _check_can_add(
            self,
            handler: IoPipelineHandler,
            *,
            name: ta.Optional[str] = None,
    ) -> IoPipelineHandler:
        check.state(self._state == IoPipeline.State.READY)  # noqa

        if not isinstance(handler, ShareableIoPipelineHandler):
            check.not_in(handler, self._unique_contexts)

        if name is not None:
            check.not_in(name, self._contexts_by_name)

        return handler

    def _check_can_add_relative_to(self, ctx: IoPipelineHandlerContext) -> IoPipelineHandlerContext:
        check.is_(ctx._pipeline, self)  # noqa
        check.state(not ctx._invalidated)  # noqa

        return ctx

    def _add(
            self,
            handler: IoPipelineHandler,
            *,
            inner_to: ta.Optional[IoPipelineHandlerContext] = None,
            outer_to: ta.Optional[IoPipelineHandlerContext] = None,

            name: ta.Optional[str] = None,
    ) -> IoPipelineHandlerRef:
        self._check_can_add(handler, name=name)

        if inner_to is not None:
            check.none(outer_to)
            check.is_not(inner_to, self._innermost)
            self._check_can_add_relative_to(inner_to)
        elif outer_to is not None:
            check.none(inner_to)
            check.is_not(outer_to, self._outermost)
            self._check_can_add_relative_to(outer_to)
        else:
            raise ValueError('Must specify exactly one of inner_to or outer_to')

        ctx = IoPipelineHandlerContext(
            _pipeline=self,
            _handler=handler,

            _name=name,
        )

        self._handler_update(ctx, 'adding')  # noqa

        if isinstance(handler, ShareableIoPipelineHandler):
            self._shareable_contexts.setdefault(handler, set()).add(ctx)
        else:
            check.not_in(handler, self._unique_contexts)  # also pre-checked by _check_can_add
            self._unique_contexts[handler] = ctx

        if name is not None:
            self._contexts_by_name[name] = ctx

        if inner_to is not None:
            prv = inner_to._next_in  # noqa
            ctx._next_out = inner_to  # noqa
            ctx._next_in = prv  # noqa
            inner_to._next_in = ctx  # noqa
            prv._next_out = ctx  # noqa

        if outer_to is not None:
            prv = outer_to._next_out  # noqa
            ctx._next_out = prv  # noqa
            ctx._next_in = outer_to  # noqa
            prv._next_in = ctx  # noqa
            outer_to._next_out = ctx  # noqa

        self._clear_caches()

        self._handler_update(ctx, 'added')  # noqa

        # FIXME: exceptions?
        self._notify(ctx, IoPipelineHandlerNotifications.Added(ctx))

        return ctx._ref  # noqa

    def add_innermost(
            self,
            handler: IoPipelineHandler,
            *,
            name: ta.Optional[str] = None,
    ) -> IoPipelineHandlerRef:
        return self._add(handler, outer_to=self._innermost, name=name)

    def add_outermost(
            self,
            handler: IoPipelineHandler,
            *,
            name: ta.Optional[str] = None,
    ) -> IoPipelineHandlerRef:
        return self._add(handler, inner_to=self._outermost, name=name)

    def add_inner_to(
            self,
            inner_to: IoPipelineHandlerRef,
            handler: IoPipelineHandler,
            *,
            name: ta.Optional[str] = None,
    ) -> IoPipelineHandlerRef:
        ctx = inner_to._context   # noqa
        return self._add(handler, inner_to=ctx, name=name)

    def add_outer_to(
            self,
            outer_to: IoPipelineHandlerRef,
            handler: IoPipelineHandler,
            *,
            name: ta.Optional[str] = None,
    ) -> IoPipelineHandlerRef:
        ctx = outer_to._context   # noqa
        return self._add(handler, outer_to=ctx, name=name)

    #

    def _check_can_remove(self, handler_ref: IoPipelineHandlerRef) -> IoPipelineHandler:
        ctx = handler_ref._context  # noqa
        check.is_(ctx._pipeline, self)  # noqa

        check.state(self._state in (IoPipeline.State.READY, IoPipeline.State.DESTROYING))  # noqa

        check.state(not ctx._invalidated)  # noqa

        handler = ctx._handler  # noqa
        if isinstance(handler, ShareableIoPipelineHandler):
            check.in_(ctx, self._shareable_contexts[handler])
        else:
            check.equal(ctx, self._unique_contexts[handler])

        check.is_not(ctx, self._innermost)
        check.is_not(ctx, self._outermost)

        return handler

    def _remove(self, handler_ref: IoPipelineHandlerRef) -> None:
        self._check_can_remove(handler_ref)

        ctx = handler_ref._context  # noqa
        handler = ctx._handler  # noqa

        self._handler_update(ctx, 'removing')  # noqa

        if ctx._name is not None:  # noqa
            del self._contexts_by_name[ctx._name]  # noqa

        if isinstance(handler, ShareableIoPipelineHandler):
            cs = self._shareable_contexts[handler]
            cs.remove(ctx)
            if not cs:
                del self._shareable_contexts[handler]
        else:
            del self._unique_contexts[handler]

        ctx._next_in._next_out = ctx._next_out  # noqa
        ctx._next_out._next_in = ctx._next_in  # noqa

        ctx._invalidated = True  # noqa
        del ctx._next_in  # noqa
        del ctx._next_out  # noqa

        self._clear_caches()

        self._handler_update(ctx, 'removed')  # noqa

        # FIXME: exceptions? defer?
        self._notify(ctx, IoPipelineHandlerNotifications.Removed(ctx))

    def remove(self, handler_ref: IoPipelineHandlerRef) -> None:
        self._remove(handler_ref)

    #

    def replace(
            self,
            old_handler_ref: IoPipelineHandlerRef,
            new_handler: IoPipelineHandler,
            *,
            name: ta.Optional[str] = None,
    ) -> IoPipelineHandlerRef:
        self._check_can_remove(old_handler_ref)
        self._check_can_add(new_handler, name=name)

        inner_to = old_handler_ref._context._next_out  # noqa
        self._remove(old_handler_ref)
        return self._add(new_handler, inner_to=inner_to, name=name)

    #

    @ta.final
    class _Outermost(IoPipelineHandler):
        """'Head' in Netty terms."""

        def __repr__(self) -> str:
            return f'{type(self).__name__}'

        def outbound(self, ctx: 'IoPipelineHandlerContext', msg: ta.Any) -> None:
            if isinstance(msg, IoPipelineMessages.MustPropagate):
                ctx._pipeline._propagation.remove_must(ctx, 'outbound', msg)  # noqa

            ctx._pipeline._terminal_outbound(ctx, msg)  # noqa

    @ta.final
    class _Innermost(IoPipelineHandler):
        """'Tail' in Netty terms."""

        def __repr__(self) -> str:
            return f'{type(self).__name__}'

        def inbound(self, ctx: 'IoPipelineHandlerContext', msg: ta.Any) -> None:
            if isinstance(msg, IoPipelineMessages.MustPropagate):
                ctx._pipeline._propagation.remove_must(ctx, 'inbound', msg)  # noqa

            ctx._pipeline._terminal_inbound(ctx, msg)  # noqa

    ##
    # caches

    @ta.final
    class _Caches:
        def __init__(self, p: 'IoPipeline') -> None:
            self._p = p

            self._handlers_by_type_cache: ta.Dict[type, ta.Sequence[IoPipelineHandlerRef]] = {}
            self._single_handlers_by_type_cache: ta.Dict[type, ta.Optional[IoPipelineHandlerRef]] = {}

        _handlers: ta.Sequence[IoPipelineHandlerRef_]

        def handlers(self) -> ta.Sequence[IoPipelineHandlerRef_]:
            try:
                return self._handlers
            except AttributeError:
                pass

            lst: ta.List[IoPipelineHandlerRef_] = []
            ctx = self._p._outermost  # noqa
            while (ctx := ctx._next_in) is not self._p._innermost:  # noqa
                lst.append(ctx._ref)  # noqa

            self._handlers = lst
            return lst

        _handlers_by_name: ta.Mapping[str, IoPipelineHandlerRef_]

        def handlers_by_name(self) -> ta.Mapping[str, IoPipelineHandlerRef_]:
            try:
                return self._handlers_by_name
            except AttributeError:
                pass

            dct: ta.Dict[str, IoPipelineHandlerRef_] = {}
            ctx = self._p._outermost  # noqa
            while (ctx := ctx._next_in) is not self._p._innermost:  # noqa
                if (n := ctx._name) is not None:  # noqa
                    dct[n] = ctx._ref  # noqa

            self._handlers_by_name = dct
            return dct

        def find_handlers_of_type(self, ty: ta.Type[T]) -> ta.Sequence[IoPipelineHandlerRef[T]]:
            try:
                return self._handlers_by_type_cache[ty]
            except KeyError:
                pass

            ret: ta.List[ta.Any] = []
            ctx = self._p._outermost  # noqa
            while (ctx := ctx._next_in) is not self._p._innermost:  # noqa
                if isinstance(ctx._handler, ty):  # noqa
                    ret.append(ctx._ref)  # noqa

            self._handlers_by_type_cache[ty] = ret
            return ret

        def find_single_handler_of_type(self, ty: ta.Type[T]) -> ta.Optional[IoPipelineHandlerRef[T]]:
            try:
                return self._single_handlers_by_type_cache[ty]
            except KeyError:
                pass

            self._single_handlers_by_type_cache[ty] = ret = check.opt_single(self.find_handlers_of_type(ty))
            return ret

    __caches: _Caches

    def _caches(self) -> _Caches:
        try:
            return self.__caches
        except AttributeError:
            pass
        self.__caches = caches = IoPipeline._Caches(self)
        return caches

    def _clear_caches(self) -> None:
        try:
            del self.__caches
        except AttributeError:
            pass

    def handlers(self) -> ta.Sequence[IoPipelineHandlerRef]:
        return self._caches().handlers()

    def handlers_by_name(self) -> ta.Mapping[str, IoPipelineHandlerRef_]:
        return self._caches().handlers_by_name()

    @dc.dataclass(frozen=True)
    class HandlerType(ta.Generic[T]):
        """This is entirely just a workaround for mypy's `type-abstract` deficiency."""

        ty: ta.Type[T]

    def find_handlers_of_type(
            self,
            ty: ta.Union[HandlerType[T], ta.Type[T]],
    ) -> ta.Sequence[IoPipelineHandlerRef[T]]:
        if isinstance(ty, IoPipeline.HandlerType):
            ty = ty.ty
        return self._caches().find_handlers_of_type(ty)

    def find_single_handler_of_type(
            self,
            ty: ta.Union[HandlerType[T], ta.Type[T]],
    ) -> ta.Optional[IoPipelineHandlerRef[T]]:
        if isinstance(ty, IoPipeline.HandlerType):
            ty = ty.ty
        return self._caches().find_single_handler_of_type(ty)

    #

    @ta.overload
    def find_handler(  # type: ignore[overload-overlap]
            self,
            handler: ShareableIoPipelineHandlerT,
    ) -> ta.Sequence[IoPipelineHandlerRef[ShareableIoPipelineHandlerT]]:
        ...

    @ta.overload
    def find_handler(
            self,
            handler: IoPipelineHandlerT,
    ) -> ta.Optional[IoPipelineHandlerRef[IoPipelineHandlerT]]:
        ...

    def find_handler(self, handler):
        if isinstance(handler, ShareableIoPipelineHandler):
            out: ta.List[ta.Any] = []
            ctx = self._outermost
            while (ctx := ctx._next_in) is not self._innermost:  # noqa
                if handler == ctx._handler:  # noqa
                    out.append(ctx._ref)  # noqa
            return out

        else:
            # Relies on existing uniqueness checks
            ctx = self._outermost
            while (ctx := ctx._next_in) is not self._innermost:  # noqa
                if handler == ctx._handler:  # noqa
                    return ctx._ref  # noqa
            return None

    ##
    # destroy

    def __enter__(self) -> 'IoPipeline':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.destroy()

    def destroy(self) -> None:
        if self._state == IoPipeline.State.DESTROYED:
            return

        check.state(self._state == IoPipeline.State.READY)
        self._state = IoPipeline.State.DESTROYING

        self._step_in()
        try:
            im_ctx = self._innermost  # noqa
            om_ctx = self._outermost  # noqa
            while (ctx := im_ctx._next_out) is not om_ctx:  # noqa
                self.remove(ctx._ref)  # noqa

        finally:
            self._step_out()

        for svc in self._services._handles_pipeline_update:  # noqa
            svc.pipeline_update(self, 'removed')

        self._state = IoPipeline.State.DESTROYED



from .base import (  # noqa
    SCOPE,

    BaseServerUrl,
    BASE_SERVER_URL,

    url_for,
)

from .markers import (  # noqa
    AppMarker,

    append_app_marker,
    get_app_markers,

    AppMarkerProcessor,
    NopAppMarkerProcessor,

    AppMarkerProcessorMap,
)

from .routes import (  # noqa
    KnownMethod,
    KNOWN_METHODS,

    Route,
    RouteHandler,

    handles,
    HANDLES_APP_MARKER_PROCESSORS,

    RouteHandlerHolder,
    get_marked_route_handlers,

    DuplicateRouteError,
    build_route_handler_map,

    RouteHandlerApp,
)

from .sessions import (  # noqa
    SESSION,
    with_session,
)



import abc
import dataclasses as dc
import importlib.resources
import typing as ta

from ... import check
from ... import lang
from ...logs import all as logs
from ...text import templating as tpl
from .base import url_for


with lang.auto_proxy_import(globals()):
    import jinja2


log = logs.get_module_logger(globals())


##


TEMPLATE_DEFAULT_NAMESPACE: ta.Mapping[str, ta.Callable] = {}


def default_template_helper(fn):
    check.not_in(n := fn.__name__, TEMPLATE_DEFAULT_NAMESPACE)
    ta.cast(ta.Any, TEMPLATE_DEFAULT_NAMESPACE)[n] = fn
    return fn


default_template_helper(url_for)


##


TemplateNamespace = ta.NewType('TemplateNamespace', ta.Mapping[str, ta.Any])


class Templates:
    @dc.dataclass(frozen=True)
    class Config:
        resource_root: str
        reload: bool = False

    def __init__(
            self,
            config: Config,
            ns: TemplateNamespace,
    ) -> None:
        super().__init__()

        self._config = config
        self._ns = ns

        self._all: ta.Mapping[str, tpl.Templater] | None = None

    @property
    @abc.abstractmethod
    def _file_extensions(self) -> ta.AbstractSet[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def _build_one(self, s: str) -> tpl.Templater:
        raise NotImplementedError

    def _load_all(self) -> ta.Mapping[str, tpl.Templater]:
        exts = self._file_extensions
        ret: dict[str, tpl.Templater] = {}

        for fn in importlib.resources.files(self._config.resource_root).iterdir():  # noqa
            if any(fn.name.endswith(ext) for ext in exts):
                ret[fn.name] = self._build_one(fn.read_text())

        return ret

    def load_all(self) -> ta.Mapping[str, tpl.Templater]:
        if self._config.reload:
            return self._load_all()

        if self._all is None:
            self._all = self._load_all()
        return self._all

    def load(self, name: str) -> tpl.Templater:
        return self.load_all()[name]

    def render(self, template_name: str, **kwargs: ta.Any) -> bytes:
        return self.load(template_name).render(tpl.Templater.Context(env={
            **TEMPLATE_DEFAULT_NAMESPACE,
            **self._ns,
            **kwargs,
        })).encode()


class JinjaTemplates(Templates):
    def __init__(
            self,
            config: Templates.Config,
            ns: TemplateNamespace,
    ) -> None:
        super().__init__(
            config,
            ns,
        )

        self._env = jinja2.Environment(
            loader=self._loader_cls()(self),
            autoescape=True,
        )

    _loader_cls_: ta.Any

    @classmethod
    def _loader_cls(cls) -> type:
        try:
            return cls._loader_cls_
        except AttributeError:
            pass

        class _Loader(jinja2.BaseLoader):  # noqa
            def __init__(self, owner: 'JinjaTemplates') -> None:
                super().__init__()

                self._owner = owner

            def get_source(self, environment, template):
                raise TypeError

            def list_templates(self):
                raise TypeError

            def load(self, environment, name, globals=None):  # noqa
                return check.isinstance(self._owner.load(name), tpl.JinjaTemplater).tmpl

        return _Loader

    @property
    def _file_extensions(self) -> ta.AbstractSet[str]:
        return {'.j2'}

    def _build_one(self, s: str) -> tpl.Templater:
        return tpl.JinjaTemplater.from_string(
            s,
            env=self._env,
        )



# ruff: noqa: UP006 UP045
# @omlish-lite
import abc
import typing as ta
import zlib

from ....io.types import BytesLike
from ....lite.abstract import Abstract
from ....lite.namespaces import NamespaceClass


##


class IoPiplineHttpCompressorCoding(Abstract):
    @abc.abstractmethod
    def compress(
            self,
            data: BytesLike,
            /,
    ) -> ta.Optional[BytesLike]:
        raise NotImplementedError

    def flush(self) -> ta.Optional[BytesLike]:
        return None

    @abc.abstractmethod
    def finish(self) -> ta.Optional[BytesLike]:
        raise NotImplementedError



IoPiplineHttpCompressorCodings = ta.Mapping[str, ta.Callable[[], IoPiplineHttpCompressorCoding]]  # ta.TypeAlias  # omlish-amalg-typing-no-move  # noqa


#


class IoPiplineHttpDecompressorCoding(Abstract):
    @abc.abstractmethod
    def decompress(
            self,
            data: BytesLike,
            max_bytes: ta.Optional[int] = None,
            /,
    ) -> ta.Optional[BytesLike]:
        raise NotImplementedError

    @abc.abstractmethod
    def unconsumed_tail(self) -> ta.Optional[BytesLike]:
        raise NotImplementedError

    @abc.abstractmethod
    def finish(self) -> ta.Optional[BytesLike]:
        raise NotImplementedError


IoPiplineHttpDecompressorCodings = ta.Mapping[str, ta.Callable[[], IoPiplineHttpDecompressorCoding]]  # ta.TypeAlias  # omlish-amalg-typing-no-move  # noqa


##


class ZlibIoPiplineHttpCompressorCoding(IoPiplineHttpCompressorCoding):
    def __init__(self, wbits: int = 16 + zlib.MAX_WBITS) -> None:
        super().__init__()

        self._z = zlib.compressobj(wbits=wbits)

    def compress(
            self,
            data: BytesLike,
            /,
    ) -> ta.Optional[BytesLike]:
        return self._z.compress(data)

    def flush(self) -> ta.Optional[BytesLike]:
        return self._z.flush(zlib.Z_SYNC_FLUSH) or None

    def finish(self) -> ta.Optional[BytesLike]:
        return self._z.flush()


class ZlibIoPiplineHttpDecompressorCoding(IoPiplineHttpDecompressorCoding):
    def __init__(self, wbits: int = 16 + zlib.MAX_WBITS) -> None:
        super().__init__()

        self._z = zlib.decompressobj(wbits)

    def decompress(
            self,
            data: BytesLike,
            max_bytes: ta.Optional[int] = None,
            /,
    ) -> ta.Optional[BytesLike]:
        return self._z.decompress(data, max_bytes or 0)

    def unconsumed_tail(self) -> ta.Optional[BytesLike]:
        return self._z.unconsumed_tail

    def finish(self) -> ta.Optional[BytesLike]:
        return self._z.flush()


##


class DefaultIoPiplineHttpCompressionCodings(NamespaceClass):
    COMPRESSOR: ta.Final[IoPiplineHttpCompressorCodings] = {
        'gzip': ZlibIoPiplineHttpCompressorCoding,
    }

    DECOMPRESSOR: ta.Final[IoPiplineHttpDecompressorCodings] = {
        'gzip': ZlibIoPiplineHttpDecompressorCoding,
    }



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - better driver impl
   - only ever call create_task at startup, never in inner loops
     - nothing ever does `asyncio.wait(...)`
   - dedicated read_task, flush_task, sched_task
     - read_task toggles back and forth between reading and waiting
   - main task only reads from command queue
 - asynclite?
"""
import abc
import asyncio
import dataclasses as dc
import functools
import typing as ta

from ....lite.abstract import Abstract
from ....lite.check import check
from ....logs.modules import get_module_loggers
from ....logs.utils import async_exception_logging
from ...streams.utils import ByteStreamBuffers
from ..asyncs import AsyncIoPipelineMessages
from ..core import IoPipeline
from ..core import IoPipelineHandlerRef
from ..core import IoPipelineMessages
from ..core import IoPipelineService
from ..core import IoPipelineServices
from ..flow.types import IoPipelineFlow
from ..flow.types import IoPipelineFlowMessages
from ..sched.types import IoPipelineScheduling
from .metadata import DriverIoPipelineMetadata


log, alog = get_module_loggers(globals())  # noqa


##


class AsyncioStreamIoPipelineDriver(Abstract):
    @dc.dataclass(frozen=True)
    class Config:
        DEFAULT: ta.ClassVar['AsyncioStreamIoPipelineDriver.Config']

        read_chunk_size: int = 64 * 1024
        write_chunk_max: ta.Optional[int] = None

    Config.DEFAULT = Config()

    #

    def __init__(
            self,
            spec: IoPipeline.Spec,
            reader: asyncio.StreamReader,
            writer: ta.Optional[asyncio.StreamWriter] = None,
            config: ta.Optional[Config] = None,
    ) -> None:
        super().__init__()

        self._spec = spec
        self._reader = reader
        self._writer = writer
        if config is None:
            config = AsyncioStreamIoPipelineDriver.Config.DEFAULT
        self._config = config

        #

        self._shutdown_event = asyncio.Event()

        self._command_queue: asyncio.Queue[AsyncioStreamIoPipelineDriver._Command] = asyncio.Queue()

    def __repr__(self) -> str:
        return f'{type(self).__name__}@{id(self):x}'

    @property
    def config(self) -> Config:
        return self._config

    @property
    def pipeline(self) -> IoPipeline:
        return self._pipeline

    ##
    # init

    _sched: 'AsyncioStreamIoPipelineDriver._SchedulingService'

    _pipeline: IoPipeline

    _flow: ta.Optional[IoPipelineFlow]

    _command_handlers: ta.Mapping[ta.Type['AsyncioStreamIoPipelineDriver._Command'], ta.Callable[[ta.Any], ta.Awaitable[None]]]  # noqa
    _output_handlers: ta.Mapping[type, ta.Callable[[ta.Any], ta.Awaitable[None]]]

    async def _init(self) -> None:
        self._sched = self._SchedulingService(self)

        services = IoPipelineServices.of(self._spec.services)
        self._flow = services.find(IoPipelineFlow)

        self._command_handlers = self._build_command_handlers()
        self._output_handlers = self._build_output_handlers()

        #

        self._pipeline = IoPipeline(dc.replace(
            self._spec,
            metadata=(*self._spec.metadata, DriverIoPipelineMetadata(self)),
            services=(*self._spec.services, self._sched),
        ))

    def _is_auto_read(self) -> bool:
        return (flow := self._flow) is None or flow.is_auto_read()

    ##
    # async utils

    @staticmethod
    async def _cancel_tasks(
            *tasks: ta.Optional[asyncio.Task],
            check_running: bool = False,
    ) -> None:
        if check_running:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                return
            else:
                if not loop.is_running():
                    return

        #

        cts: ta.List[asyncio.Task] = []

        for t in tasks:
            if t is not None and not t.done():
                t.cancel()
                cts.append(t)

        if cts:
            await asyncio.gather(*cts, return_exceptions=True)

    #

    async def _close_writer(self) -> None:
        if self._writer is None:
            return

        try:
            self._writer.close()
            await self._writer.wait_closed()

        except Exception:  # noqa
            # Best effort; transport close errors aren't actionable at this layer.
            pass

        self._writer = None

    ##

    class _Command(Abstract):
        pass

    ##
    # feed in

    @dc.dataclass(frozen=True)
    class _FeedInCommand(_Command):
        msgs: ta.Sequence[ta.Any]

        fut: ta.Optional['asyncio.Future[None]'] = None

        def __repr__(self) -> str:
            return f'{self.__class__.__name__}([{", ".join(map(repr, self.msgs))}])'

    async def _handle_command_feed_in(self, cmd: _FeedInCommand) -> None:
        async def _inner() -> None:
            self._pipeline.feed_in(*cmd.msgs)  # noqa

        if (fut := cmd.fut) is None:
            await self._do_with_pipeline(_inner)
            return

        try:
            await self._do_with_pipeline(_inner)
            fut.set_result(None)
        except BaseException as e:  # noqa
            fut.set_exception(e)
            raise

    def enqueue_waitable(self, *msgs: ta.Any) -> 'asyncio.Future[None]':
        check.state(not self._shutdown_event.is_set())

        fut: asyncio.Future[None] = asyncio.Future()
        self._command_queue.put_nowait(AsyncioStreamIoPipelineDriver._FeedInCommand(msgs, fut=fut))
        return fut

    def enqueue(self, *msgs: ta.Any) -> None:
        check.state(not self._shutdown_event.is_set())

        self._command_queue.put_nowait(AsyncioStreamIoPipelineDriver._FeedInCommand(msgs))

    ##
    # read completed

    class _ReadCompletedCommand(_Command):
        def __init__(self, data: ta.Union[bytes, ta.List[bytes]]) -> None:
            self._data = data

        def __repr__(self) -> str:
            return (
                f'{self.__class__.__name__}@{id(self):x}'
                f'({"[...]" if isinstance(self._data, list) else "..." if self._data is not None else ""})'
            )

        def data(self) -> ta.Sequence[bytes]:
            if isinstance(self._data, bytes):
                return [self._data]
            elif isinstance(self._data, list):
                return self._data
            else:
                raise TypeError(self._data)

    class _ReadCancelledCommand(_Command):
        pass

    _pending_completed_reads: ta.Optional[ta.List[_ReadCompletedCommand]] = None

    async def _handle_command_read_completed(self, cmd: _ReadCompletedCommand) -> None:
        if not (self._want_read or self._is_auto_read()):
            if (pl := self._pending_completed_reads) is None:
                pl = self._pending_completed_reads = []

            pl.append(cmd)
            return

        #

        eof = False

        in_msgs: ta.List[ta.Any] = []

        for b in cmd.data():
            check.state(not eof)
            if not b:
                eof = True
            else:
                in_msgs.append(b)

        if self._flow is not None:
            in_msgs.append(IoPipelineFlowMessages.FlushInput())

        if eof:
            self._has_read_eof = True

            in_msgs.append(IoPipelineMessages.FinalInput())

        #

        async def _inner() -> None:
            self._pipeline.feed_in(*in_msgs)

        await self._do_with_pipeline(_inner)

        #

        if eof:
            self._shutdown_event.set()

            await self._close_writer()

    ##
    # update want read

    class _UpdateWantReadCommand(_Command):
        pass

    _has_sent_update_want_read_command: bool = False

    async def _send_update_want_read_command(self) -> None:
        if self._has_sent_update_want_read_command:
            return

        self._has_sent_update_want_read_command = True
        await self._command_queue.put(AsyncioStreamIoPipelineDriver._UpdateWantReadCommand())

    _want_read = False

    _delay_sending_update_want_read_command = False

    async def _set_want_read(self, want_read: bool) -> None:
        if self._want_read == want_read:
            return

        self._want_read = want_read

        if not self._delay_sending_update_want_read_command:
            await self._send_update_want_read_command()

    async def _handle_command_update_want_read(self, cmd: _UpdateWantReadCommand) -> None:
        self._sent_update_want_read_command = False

        if self._want_read:
            if self._pending_completed_reads:
                in_cmd = AsyncioStreamIoPipelineDriver._ReadCompletedCommand([
                    b
                    for pcr_cmd in self._pending_completed_reads
                    for b in pcr_cmd.data()
                ])
                self._command_queue.put_nowait(in_cmd)
                self._pending_completed_reads = None

            self._ensure_read_task()

    _has_read_eof = False

    def _maybe_ensure_read_task(self) -> None:
        if not self._has_read_eof and (self._want_read or self._is_auto_read()):
            self._ensure_read_task()

    @abc.abstractmethod
    def _ensure_read_task(self) -> None:
        raise NotImplementedError

    ##
    # scheduling

    class _SchedulingService(IoPipelineScheduling, IoPipelineService):
        def __init__(self, d: 'AsyncioStreamIoPipelineDriver') -> None:
            super().__init__()

            self._d = d

            self._pending: ta.List[ta.Tuple[float, ta.Callable[[], None]]] = []
            self._tasks: ta.Set[asyncio.Task] = set()

        class _Handle(IoPipelineScheduling.Handle):
            def cancel(self) -> None:
                raise NotImplementedError

        def schedule(
                self,
                handler_ref: IoPipelineHandlerRef,
                delay_s: float,
                fn: ta.Callable[[], None],
        ) -> IoPipelineScheduling.Handle:
            self._pending.append((delay_s, fn))
            return self._Handle()

        def cancel_all(self, handler_ref: ta.Optional[IoPipelineHandlerRef] = None) -> None:
            raise NotImplementedError

        async def _task_body(self, delay: float, fn: ta.Callable[[], None]) -> None:
            await asyncio.sleep(delay)

            self._d._command_queue.put_nowait(AsyncioStreamIoPipelineDriver._ScheduledCommand(fn))  # noqa

        async def _flush_pending(self) -> None:
            if not (lst := self._pending):
                return

            for delay, fn in lst:
                self._tasks.add(asyncio.create_task(functools.partial(self._task_body, delay, fn)()))

            self._pending = []

    @dc.dataclass(frozen=True)
    class _ScheduledCommand(_Command):
        fn: ta.Callable[[], None]

    async def _handle_scheduled_command(self, cmd: _ScheduledCommand) -> None:
        async def _inner() -> None:
            with self._pipeline.enter():
                cmd.fn()

        await self._do_with_pipeline(_inner)

    # handlers

    def _build_command_handlers(self) -> ta.Mapping[ta.Type[_Command], ta.Callable[[ta.Any], ta.Awaitable[None]]]:
        return {
            AsyncioStreamIoPipelineDriver._FeedInCommand: self._handle_command_feed_in,
            AsyncioStreamIoPipelineDriver._ReadCompletedCommand: self._handle_command_read_completed,
            AsyncioStreamIoPipelineDriver._UpdateWantReadCommand: self._handle_command_update_want_read,
            AsyncioStreamIoPipelineDriver._ScheduledCommand: self._handle_scheduled_command,
        }

    async def _handle_command(self, cmd: _Command) -> None:
        log.debug(lambda: f'Handling command: {cmd!r}')

        try:
            fn = self._command_handlers[cmd.__class__]
        except KeyError:
            raise TypeError(f'Unknown command type: {cmd.__class__}') from None

        await fn(cmd)

    ##
    # output handling

    # lifecycle

    async def _handle_output_final_output(self, msg: IoPipelineMessages.FinalOutput) -> None:
        self._shutdown_event.set()

        await self._close_writer()

    # defer

    async def _handle_output_defer(self, msg: IoPipelineMessages.Defer) -> None:
        self._pipeline.run_deferred(msg)

    # data (special cased)

    async def _handle_output_bytes(self, msg: ta.Any) -> None:
        for mv in ByteStreamBuffers.iter_segments(msg):
            if self._writer is not None and mv:
                self._writer.write(mv)

    # flow

    async def _handle_output_flush_output(self, msg: IoPipelineFlowMessages.FlushOutput) -> None:
        if self._writer is not None:
            await self._writer.drain()

    async def _handle_output_ready_for_input(self, msg: IoPipelineFlowMessages.ReadyForInput) -> None:
        await self._set_want_read(True)

    # async

    async def _handle_output_await(self, msg: AsyncIoPipelineMessages.Await) -> None:
        try:
            result = await msg.obj

        except Exception as e:  # noqa
            with self._pipeline.enter():
                msg.set_failed(e)

        else:
            with self._pipeline.enter():
                msg.set_succeeded(result)

    # handlers

    def _build_output_handlers(self) -> ta.Mapping[type, ta.Callable[[ta.Any], ta.Awaitable[None]]]:
        return {
            IoPipelineMessages.FinalOutput: self._handle_output_final_output,
            IoPipelineMessages.Defer: self._handle_output_defer,
            IoPipelineFlowMessages.FlushOutput: self._handle_output_flush_output,
            IoPipelineFlowMessages.ReadyForInput: self._handle_output_ready_for_input,
            AsyncIoPipelineMessages.Await: self._handle_output_await,
        }

    async def _handle_output(self, msg: ta.Any) -> None:
        log.debug(lambda: f'Handling output: {msg!r}')

        if ByteStreamBuffers.can_bytes(msg):
            await self._handle_output_bytes(msg)
            return

        try:
            fn = self._output_handlers[msg.__class__]
        except KeyError:
            raise TypeError(f'Unknown output type: {msg.__class__}') from None

        await fn(msg)

    # execution helpers

    async def _do_with_pipeline(self, fn: ta.Callable[[], ta.Awaitable[None]]) -> None:
        prev_want_read = self._want_read
        if not self._is_auto_read():
            self._want_read = False

        self._delay_sending_update_want_read_command = True
        try:
            await fn()

            await self._drain_pipeline_output()

        finally:
            self._delay_sending_update_want_read_command = False

        if self._shutdown_event.is_set():
            return

        await self._sched._flush_pending()  # noqa

        if self._want_read != prev_want_read:
            await self._send_update_want_read_command()

        self._maybe_ensure_read_task()

    async def _drain_pipeline_output(self) -> None:
        while (msg := self._pipeline.output.poll()) is not None:
            await self._handle_output(msg)

    ##
    # shutdown

    _shutdown_task: asyncio.Task

    async def _shutdown_task_main(self) -> None:
        await self._shutdown_event.wait()

    ##
    # main loop

    @abc.abstractmethod
    def _run(self) -> ta.Awaitable[None]:
        raise NotImplementedError

    @async_exception_logging(alog)
    async def run(self) -> None:
        try:
            self._shutdown_task  # noqa
        except AttributeError:
            pass
        else:
            raise RuntimeError('Already running')

        await self._init()

        self._shutdown_task = asyncio.create_task(self._shutdown_task_main())

        try:
            try:
                await self._run()

            finally:
                self._pipeline.destroy()

        finally:
            await self._cancel_tasks(self._shutdown_task, check_running=True)


##


class SimpleAsyncioStreamIoPipelineDriver(AsyncioStreamIoPipelineDriver):
    _read_task: ta.Optional[asyncio.Task] = None

    def _ensure_read_task(self) -> None:
        if self._read_task is not None or self._shutdown_event.is_set():
            return

        self._read_task = asyncio.create_task(self._reader.read(self._config.read_chunk_size))

        def _done(task: 'asyncio.Task[bytes]') -> None:
            check.state(task is self._read_task)
            self._read_task = None

            if self._shutdown_event.is_set():
                return

            cmd: AsyncioStreamIoPipelineDriver._Command
            eof = False
            try:
                data = task.result()

            except asyncio.CancelledError:
                cmd = AsyncioStreamIoPipelineDriver._ReadCancelledCommand()  # noqa

            else:
                cmd = AsyncioStreamIoPipelineDriver._ReadCompletedCommand(data)  # noqa
                eof = not data

            self._command_queue.put_nowait(cmd)

            # FIXME: disable? toggle? this pre-reads till told to stop to try to reduce stalling
            if not eof:
                self._maybe_ensure_read_task()

        self._read_task.add_done_callback(_done)

    #

    async def _run(self) -> None:
        self._maybe_ensure_read_task()

        #

        command_queue_task: ta.Optional[asyncio.Task[AsyncioStreamIoPipelineDriver._Command]] = None

        try:
            if not self._shutdown_event.is_set():
                await self._handle_command(AsyncioStreamIoPipelineDriver._FeedInCommand([  # noqa
                    IoPipelineMessages.InitialInput(),
                ]))

            while not self._shutdown_event.is_set():
                if command_queue_task is None:
                    command_queue_task = asyncio.create_task(self._command_queue.get())

                done, pending = await asyncio.wait(
                    [command_queue_task, self._shutdown_task],
                    return_when=asyncio.FIRST_COMPLETED,
                )

                winner = done.pop()

                if self._shutdown_event.is_set() or winner is self._shutdown_task:
                    break

                elif winner is command_queue_task:
                    cmd = command_queue_task.result()
                    command_queue_task = None

                    await self._handle_command(cmd)

                    del cmd
                    command_queue_task = None

                else:
                    raise RuntimeError(f'Unexpected task: {winner!r}')
        #

        finally:
            await self._cancel_tasks(
                command_queue_task,
                self._read_task,
                check_running=True,
            )



# ruff: noqa: FURB188 UP045
# @omlish-lite
import typing as ta

from ...streams.utils import ByteStreamBuffers
from ..core import IoPipelineHandlerFn
from ..handlers.fns import IoPipelineHandlerFns
from ..handlers.queues import InboundQueueIoPipelineHandler
from .buffering import InboundBytesBufferingIoPipelineHandler


##


class InboundBytesBufferingQueueIoPipelineHandler(
    InboundBytesBufferingIoPipelineHandler,
    InboundQueueIoPipelineHandler,
):
    def __init__(
            self,
            *,
            filter: ta.Union[IoPipelineHandlerFn[ta.Any, bool], ta.Literal[True], None] = None,  # noqa
            passthrough: bool = False,
    ) -> None:
        if filter is True:
            filter = IoPipelineHandlerFns.no_context(ByteStreamBuffers.can_bytes)  # noqa

        super().__init__(
            filter=filter,  # noqa
            passthrough=passthrough,
        )

    _buffered_bytes: int = 0

    # @ta.override
    def inbound_buffered_bytes(self) -> ta.Optional[int]:
        return self._buffered_bytes

    # @ta.override
    def _append(self, msg: ta.Any) -> None:
        bl = ByteStreamBuffers.bytes_len(msg, True)

        super()._append((msg, bl))

        if bl is not None:
            self._buffered_bytes += bl

    # @ta.override
    def _popleft(self) -> ta.Any:
        msg, bl = self._q.popleft()

        if bl is not None:
            self._buffered_bytes -= bl

        return msg



# @omlish-lite
import abc
import typing as ta

from ...lite.abstract import Abstract


##


class AsyncliteIdentities(Abstract):
    """
    Get the identity of the current task/thread.

    Returns:
     - asyncio: asyncio.Task | None (from asyncio.current_task())
     - sync: threading.Thread (from threading.current_thread())
     - anyio: anyio.TaskInfo (from anyio.get_current_task())
    """

    @abc.abstractmethod
    def current_identity(self) -> ta.Any:
        raise NotImplementedError



r"""
TODO:
 - gson switches
  - Strictness.LEGACY_STRICT
   - JsonReader allows the literals true, false and null to have any capitalization, for example fAlSe or NULL
   - JsonReader supports the escape sequence \', representing a ' (single-quote)
   - JsonReader supports the escape sequence \LF (with LF being the Unicode character U+000A), resulting in a LF within
     the read JSON string
   - JsonReader allows unescaped control characters (U+0000 through U+001F)
  - Strictness.LENIENT
   - In lenient mode, all input that is accepted in legacy strict mode is accepted in addition to the following
     departures from RFC 8259:
   - Streams that start with the non-execute prefix, ")]'\n"}
   - Streams that include multiple top-level values. With legacy strict or strict parsing, each stream must contain
     exactly one top-level value.
   - Numbers may be NaNs or infinities represented by NaN and (-)Infinity respectively.
   - End of line comments starting with // or # and ending with a newline character.
   - C-style comments starting with /* and ending with */. Such comments may not be nested.
   - Names that are unquoted or 'single quoted'.
   - Strings that are unquoted or 'single quoted'.
   - Array elements separated by ; instead of ,.
   - Unnecessary array separators. These are interpreted as if null was the omitted value.
   - Names and values separated by = or => instead of :.
   - Name/value pairs separated by ; instead of ,.
"""
import itertools
import typing as ta

from .... import lang
from .building import JsonValueBuilder
from .errors import JsonStreamError
from .lexing import JsonStreamLexer
from .lexing import Token
from .parsing import Event
from .parsing import JsonStreamParser


##


class JsonStreamValueParser(lang.ExitStacked):
    class Machinery(ta.NamedTuple):
        lex: JsonStreamLexer
        parse: JsonStreamParser
        build: JsonValueBuilder

    def __init__(self, m: Machinery) -> None:
        super().__init__()

        self._m = m

    #

    def _enter_contexts(self) -> None:
        self._enter_context(self._m.lex)
        self._enter_context(self._m.parse)

    def feed(self, i: ta.Iterable[str]) -> ta.Iterator[ta.Any]:
        for c in i:
            for t in self._m.lex(c):
                for e in self._m.parse(t):
                    for v in self._m.build(e):  # noqa
                        yield v

    #

    @classmethod
    def parse_values(
            cls,
            m: Machinery,
            i: ta.Iterable[str],
    ) -> ta.Iterator[ta.Any]:
        with cls(m) as p:
            yield from p.feed(itertools.chain(i, ['']))

    @classmethod
    def parse_one_value(
            cls,
            m: Machinery,
            i: ta.Iterable[str],
    ) -> ta.Any:
        with cls(m) as p:
            return next(p.feed(itertools.chain(i, [''])))

    @classmethod
    def parse_exactly_one_value(
            cls,
            m: Machinery,
            i: ta.Iterable[str],
    ) -> ta.Any:
        r: ta.Any
        r = not_set = object()
        with cls(m) as p:
            for v in p.feed(itertools.chain(i, [''])):
                if r is not_set:
                    r = v
                else:
                    raise JsonStreamError('Unexpected input')
        if r is not_set:
            raise JsonStreamError('No value')
        return r


##


class DebugJsonStreamValueParser(JsonStreamValueParser):
    def __init__(self, m: JsonStreamValueParser.Machinery) -> None:
        super().__init__(m)

        self._chars: list[str] = []
        self._tokens: list[Token] = []
        self._events: list[Event] = []
        self._values: list[ta.Any] = []

    def feed(self, i: ta.Iterable[str]) -> ta.Iterator[ta.Any]:
        for c in i:
            self._chars.append(c)
            for t in self._m.lex(c):
                self._tokens.append(t)
                for e in self._m.parse(t):
                    self._events.append(e)
                    for v in self._m.build(e):
                        self._values.append(v)
                        yield v


##


def make_machinery(
        *,
        include_raw: bool = False,
        yield_object_lists: bool = False,
) -> JsonStreamValueParser.Machinery:
    return JsonStreamValueParser.Machinery(
        JsonStreamLexer(
            include_raw=include_raw,
        ),

        JsonStreamParser(),

        JsonValueBuilder(
            yield_object_lists=yield_object_lists,
        ),
    )


def stream_parse_values(i: ta.Iterable[str], **kwargs: ta.Any) -> ta.Iterator[ta.Any]:
    return JsonStreamValueParser.parse_values(make_machinery(**kwargs), i)


def stream_parse_one_value(i: ta.Iterable[str], **kwargs: ta.Any) -> ta.Any:
    return JsonStreamValueParser.parse_one_value(make_machinery(**kwargs), i)


def stream_parse_exactly_one_value(i: ta.Iterable[str], **kwargs: ta.Any) -> ta.Any:
    return JsonStreamValueParser.parse_exactly_one_value(make_machinery(**kwargs), i)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import typing as ta

from ..types import BytesLike
from .base import BaseByteStreamBufferLike
from .direct import _EMPTY_DIRECT_BYTE_STREAM_BUFFER_VIEW
from .direct import DirectByteStreamBufferView
from .errors import BufferTooLargeByteStreamBufferError
from .errors import NoOutstandingReserveByteStreamBufferError
from .errors import OutstandingReserveByteStreamBufferError
from .types import ByteStreamBufferView
from .types import MutableByteStreamBuffer
from .utils import ByteStreamBuffers


##


class LinearByteStreamBuffer(BaseByteStreamBufferLike, MutableByteStreamBuffer):
    """
    A simple contiguous (bytearray-backed) MutableByteStreamBuffer implementation.

    Strengths:
      - Fast `find/rfind` and contiguous peeking.
      - Efficient reserve/commit (safe against exported-memoryview resizing hazards).

    Tradeoffs:
      - `split_to` returns a stable view by copying the split bytes into an owned `bytes` object. (A truly zero-copy
        split view would require pinning the underlying bytearray against compaction.)
      - Compaction is best-effort: if the backing bytearray is pinned by exported memoryviews, compaction will be
        skipped and the buffer will remain correct but may retain memory.
    """

    def __init__(
            self,
            *,
            max_size: ta.Optional[int] = None,
            initial_capacity: int = 0,
            compact_threshold: int = 64 * 1024,
    ) -> None:
        super().__init__()

        self._max_size = None if max_size is None else int(max_size)

        if initial_capacity < 0:
            raise ValueError(initial_capacity)
        if self._max_size is not None and initial_capacity > self._max_size:
            raise BufferTooLargeByteStreamBufferError('buffer exceeded max_size')

        if compact_threshold < 0:
            raise ValueError(compact_threshold)
        self._compact_threshold = int(compact_threshold)

        # Diagnostics (best-effort compaction behavior)
        self.compaction_attempts = 0
        self.compaction_successes = 0
        self.compaction_skipped_exports = 0

        # Pre-size the backing store to encourage fewer resizes/copies on trickle-y writes. We immediately clear so
        # readable length remains 0.
        if initial_capacity:
            self._ba = bytearray(initial_capacity)
            self._ba.clear()
        else:
            self._ba = bytearray()

    @property
    def max_size(self) -> ta.Optional[int]:
        return self._max_size

    _rpos = 0
    _wpos = 0

    _resv_start: ta.Optional[int] = None
    _resv_len = 0

    _resv_buf: bytearray

    def __len__(self) -> int:
        return self._wpos - self._rpos

    def peek(self) -> memoryview:
        if self._rpos == self._wpos:
            return memoryview(b'')
        return memoryview(self._ba)[self._rpos:self._wpos]

    def segments(self) -> ta.Sequence[memoryview]:
        mv = self.peek()
        return (mv,) if len(mv) else ()

    def _check_no_reserve(self) -> None:
        if self._resv_start is not None:
            raise OutstandingReserveByteStreamBufferError('outstanding reserve')

    def write(self, data: BytesLike, /) -> None:
        self._check_no_reserve()
        if not data:
            return
        if isinstance(data, memoryview):
            data = ByteStreamBuffers.memoryview_to_bytes(data)  # noqa
        elif isinstance(data, bytearray):
            data = bytes(data)

        bl = len(data)

        if self._max_size is not None and len(self) + bl > self._max_size:
            raise BufferTooLargeByteStreamBufferError('buffer exceeded max_size')

        # If buffer is logically empty but backing store might still be large (e.g. compaction skipped),
        # keep indices reset and just append.
        if self._rpos == self._wpos:
            self._rpos = 0
            self._wpos = 0
            # Best-effort shrink on empty: may be pinned, so ignore failure.
            self._try_clear_if_empty()

        self._ba.extend(data)
        self._wpos += bl

    def reserve(self, n: int, /) -> memoryview:
        if n < 0:
            raise ValueError(n)
        if self._resv_start is not None:
            raise OutstandingReserveByteStreamBufferError('outstanding reserve')

        # Do NOT reserve by extending the backing bytearray and returning a view into it: a live exported memoryview
        # pins the bytearray against resizing, and commit() would need to shrink unused reservation (or otherwise
        # reshape), which would raise BufferError.
        #
        # Instead, reserve returns a view of a temporary bytearray, and commit() appends only what was written.
        b = bytearray(n)
        self._resv_start = 0
        self._resv_len = n
        self._resv_buf = b
        return memoryview(b)

    def commit(self, n: int, /) -> None:
        if self._resv_start is None:
            raise NoOutstandingReserveByteStreamBufferError('no outstanding reserve')
        if n < 0 or n > self._resv_len:
            raise ValueError(n)

        b = self._resv_buf
        self._resv_start = None
        self._resv_len = 0
        del self._resv_buf

        if not n:
            return

        if self._max_size is not None and len(self) + n > self._max_size:
            raise BufferTooLargeByteStreamBufferError('buffer exceeded max_size')

        # Append only what was written.
        self.write(memoryview(b)[:n])

    def advance(self, n: int, /) -> None:
        self._check_no_reserve()
        if n < 0 or n > len(self):
            raise ValueError(n)
        if not n:
            return

        self._rpos += n

        # Best-effort compaction (may skip if backing store is pinned by exported memoryviews).
        self.compact()

        # Fully consumed: reset indices and attempt to clear backing store best-effort.
        if self._rpos == self._wpos:
            self._rpos = 0
            self._wpos = 0
            self._try_clear_if_empty()

    def split_to(self, n: int, /) -> ByteStreamBufferView:
        self._check_no_reserve()
        if n < 0 or n > len(self):
            raise ValueError(n)
        if not n:
            return _EMPTY_DIRECT_BYTE_STREAM_BUFFER_VIEW

        # Copy out the split prefix to keep the view stable even if the underlying buffer compacts.
        b = ByteStreamBuffers.memoryview_to_bytes(memoryview(self._ba)[self._rpos:self._rpos + n])
        self._rpos += n

        # If we consumed everything, reset best-effort; otherwise allow compaction policy to handle later.
        if self._rpos == self._wpos:
            self._rpos = 0
            self._wpos = 0
            self._try_clear_if_empty()

        return DirectByteStreamBufferView(memoryview(b))

    def find(self, sub: bytes, start: int = 0, end: ta.Optional[int] = None) -> int:
        start, end = self._norm_slice(start, end)
        if not sub:
            return start
        i = self._ba.find(sub, self._rpos + start, self._rpos + end)
        return -1 if i < 0 else (i - self._rpos)

    def rfind(self, sub: bytes, start: int = 0, end: ta.Optional[int] = None) -> int:
        start, end = self._norm_slice(start, end)
        if not sub:
            return end
        i = self._ba.rfind(sub, self._rpos + start, self._rpos + end)
        return -1 if i < 0 else (i - self._rpos)

    def coalesce(self, n: int, /) -> memoryview:
        self._check_no_reserve()
        if n < 0:
            raise ValueError(n)
        if n > len(self):
            raise ValueError(n)
        if not n:
            return memoryview(b'')
        # Always contiguous for the readable prefix.
        return memoryview(self._ba)[self._rpos:self._rpos + n]

    def compact(
            self,
            *,
            threshold: ta.Optional[int] = None,
            min_fraction: float = 0.5,
            force: bool = False,
    ) -> bool:
        """
        Best-effort compaction to avoid "huge buffer pinned by small tail".

        Returns True if compaction happened, False otherwise.

        Compaction attempts to drop the consumed prefix of the backing bytearray by resizing it. If the bytearray is
        pinned by exported memoryviews, Python raises BufferError; in that case compaction is skipped (buffer remains
        correct) and the skip is recorded in diagnostics.

        Args:
          threshold: minimum consumed prefix size (bytes) required before compaction is considered. Defaults to the
                     instance's configured compact threshold (64KiB by default).
          min_fraction: require consumed prefix to be at least this fraction of the backing 'written' size. Defaults to
                        0.5 (i.e., consumed >= half of written region).
          force: if True, ignore heuristic checks and attempt compaction whenever `_rpos > 0`.

        Notes:
          - Compaction is disallowed while a reserve is outstanding.
          - Compaction is an optimization only; semantics do not depend on it.
        """

        self._check_no_reserve()

        if threshold is None:
            threshold = self._compact_threshold
        else:
            threshold = int(threshold)

        if not force:
            if self._rpos <= 0:
                return False
            if self._rpos < threshold:
                return False
            if self._wpos <= 0:
                return False
            if self._rpos < int(self._wpos * float(min_fraction)):
                return False
        else:
            if self._rpos <= 0:
                return False

        self.compaction_attempts += 1

        try:
            del self._ba[:self._rpos]
        except BufferError:
            self.compaction_skipped_exports += 1
            return False

        self._wpos -= self._rpos
        self._rpos = 0
        self.compaction_successes += 1
        return True

    def _try_clear_if_empty(self) -> None:
        """
        Best-effort clear of backing store when logically empty.

        If the backing store is pinned by exported memoryviews, clear will raise BufferError.
        In that case we keep the backing store allocated (correctness preserved) and move on.
        """

        if self._rpos != 0 or self._wpos != 0:
            return
        if not self._ba:
            return

        try:
            self._ba.clear()
        except BufferError:
            # Keep storage; callers wanted deterministic correctness over reclamation.
            pass



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import io
import typing as ta

from ...io.pipelines.core import IoPipelineHandler
from ...io.pipelines.core import IoPipelineHandlerContext
from ...lite.abstract import Abstract
from ..headers import HttpHeaders
from .objects import FullIoPipelineHttpMessage
from .objects import IoPipelineHttpMessageBodyData
from .objects import IoPipelineHttpMessageChunk
from .objects import IoPipelineHttpMessageChunkedTrailers
from .objects import IoPipelineHttpMessageEnd
from .objects import IoPipelineHttpMessageEndChunk
from .objects import IoPipelineHttpMessageHead
from .objects import IoPipelineHttpMessageLastChunk
from .objects import IoPipelineHttpMessageObjects


##


class IoPipelineHttpObjectEncoder(
    IoPipelineHttpMessageObjects,
    IoPipelineHandler,
    Abstract,
):
    def __init__(self) -> None:
        super().__init__()

        self._streaming = False

        self._outbound_dispatch_dct: ta.Mapping[type, ta.Callable[[IoPipelineHandlerContext, ta.Any], None]] = {
            self._head_type: self._handle_request_head,
            self._full_type: self._handle_full_request,
            self._chunk_type: self._handle_chunk,
            self._end_chunk_type: self._handle_end_chunk,
            self._last_chunk_type: self._handle_last_chunk,
            self._chunked_trailers_type: self._handle_chunked_trailers,
            self._body_data_type: self._handle_body_data,
            self._end_type: self._handle_request_end,
        }

    #

    @abc.abstractmethod
    def _encode_head_line(self, head: IoPipelineHttpMessageHead) -> bytes:
        raise NotImplementedError

    #

    def outbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if (fn := self._outbound_dispatch_dct.get(type(msg))) is not None:
            fn(ctx, msg)
            return

        ctx.feed_out(msg)

    #

    def _handle_request_head(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageHead) -> None:  # noqa
        self._streaming = True

        ctx.feed_out(self._encode_head(msg))

    #

    def _handle_full_request(self, ctx: IoPipelineHandlerContext, msg: FullIoPipelineHttpMessage) -> ta.Any:
        ctx.feed_out(self._encode_head(msg.head))
        if len(msg.body) > 0:
            ctx.feed_out(msg.body)

    #

    def _handle_chunk(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageChunk) -> None:  # noqa
        ctx.feed_out(f'{msg.size:x}\r\n'.encode('ascii'))

    #

    def _handle_end_chunk(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageEndChunk) -> None:  # noqa
        ctx.feed_out(b'\r\n')

    #

    def _handle_last_chunk(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageLastChunk) -> None:  # noqa
        ctx.feed_out(b'0\r\n')

    #

    def _handle_chunked_trailers(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageChunkedTrailers) -> None:  # noqa
        ctx.feed_out(b'\r\n')

    #

    def _handle_body_data(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageBodyData) -> None:  # noqa
        if not self._streaming:
            # Not in streaming mode - pass through unchanged
            ctx.feed_out(msg)

        if len(msg.data) < 1:
            pass

        ctx.feed_out(msg.data)

    #

    def _handle_request_end(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageEnd) -> None:  # noqa
        if not self._streaming:
            # Not in streaming mode - pass through
            ctx.feed_out(msg)
            return

        # Reset state
        self._streaming = False

    #

    def _encode_head(self, head: IoPipelineHttpMessageHead) -> bytes:
        buf = io.BytesIO()

        buf.write(self._encode_head_line(head))

        for hl in self._encode_headers(head.headers):
            buf.write(hl)

        buf.write(b'\r\n')

        return buf.getvalue()

    def _encode_headers(self, headers: HttpHeaders) -> ta.List[bytes]:
        lines: ta.List[bytes] = []

        # HttpHeaders stores entries as list of (name, value) tuples
        for name, value in headers.raw:
            # Header names and values should be ASCII-safe in practice
            line = f'{name}: {value}\r\n'.encode('ascii')
            lines.append(line)

        return lines



# @omlish-lite
import abc
import dataclasses as dc
import typing as ta

from ....lite.abstract import Abstract
from ....lite.namespaces import NamespaceClass
from ..core import IoPipeline
from ..core import IoPipelineHandlerContext
from ..core import IoPipelineMessages


##


class IoPipelineFlowMessages(NamespaceClass):
    """
    Note: these inbound messages will never be sent without a `IoPipelineFlow` instance in `channel.services` -
    thus it's safe to refer to `ctx.services[IoPipelineFlow]` when handling these.
    """

    #

    @ta.final
    @dc.dataclass(frozen=True)
    class FlushInput(  # ~ Netty `ChannelInboundInvoker::fireChannelReadComplete`  # noqa
        IoPipelineMessages.MayPropagate,
        IoPipelineMessages.NeverOutbound,
    ):
        pass

    #

    @ta.final
    @dc.dataclass(frozen=True)
    class FlushOutput(  # ~ Netty 'ChannelOutboundInvoker::flush'
        IoPipelineMessages.MayPropagate,
        IoPipelineMessages.NeverInbound,
    ):
        pass

    @ta.final
    @dc.dataclass(frozen=True)
    class ReadyForInput(  # ~ Netty `ChannelOutboundInvoker::read`
        IoPipelineMessages.MayPropagate,
        IoPipelineMessages.NeverInbound,
    ):
        pass

    #

    # # TODO:
    # @ta.final
    # @dc.dataclass(frozen=True)
    # class ReadyForOutput(  # ~ Netty `ChannelOutboundInvoker::fireChannelWritabilityChanged`  # noqa
    #     IoPipelineMessages.MayPropagate,
    #     IoPipelineMessages.NeverOutbound,
    # ):
    #     pass

    # # TODO:
    # @ta.final
    # @dc.dataclass(frozen=True)
    # class PauseOutput(  # ~ Netty `ChannelOutboundInvoker::fireChannelWritabilityChanged`  # noqa
    #     IoPipelineMessages.MayPropagate,
    #     IoPipelineMessages.NeverOutbound,
    # ):
    #     pass


##


class IoPipelineFlow(Abstract):
    @abc.abstractmethod
    def is_auto_read(
            self: ta.Union[
                'IoPipelineFlow',
                IoPipeline,
                IoPipelineHandlerContext,
                None,
            ],
    ) -> bool:
        # This strange construct grants the ability to do `IoPipelineFlow.is_auto_read(opt_flow)`, which is becoming
        # increasingly frequently useful in real code.
        if self is None:
            return False

        if isinstance(self, IoPipelineFlow):
            return self.is_auto_read()

        if isinstance(self, IoPipelineHandlerContext):
            self = self._pipeline  # noqa

        if isinstance(self, IoPipeline):
            return (fc := self.services.find(IoPipelineFlow)) is None or fc.is_auto_read()

        raise TypeError(self)



import enum
import typing as ta

from ... import check
from ... import lang
from .errors import AbnfError
from .errors import AbnfIncompleteParseError
from .matches import Match
from .matches import longest_match
from .ops import Op


with lang.auto_proxy_import(globals()):
    from . import parsing


##


class Channel(enum.Enum):
    STRUCTURE = enum.auto()
    CONTENT = enum.auto()
    COMMENT = enum.auto()
    SPACE = enum.auto()


class Rule(lang.Final):
    def __init__(
            self,
            name: str,
            op: Op,
            *,
            channel: Channel = Channel.STRUCTURE,
    ) -> None:
        super().__init__()

        self._name = check.non_empty_str(name)
        self._op = check.isinstance(op, Op)
        self._channel = channel

        self._name_f = name.casefold()

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._name!r}, channel={self._channel.name})'

    def replace_op(self, op: Op) -> 'Rule':
        return Rule(
            self._name,
            op,
            channel=self._channel,
        )

    @property
    def name(self) -> str:
        return self._name

    @property
    def name_f(self) -> str:
        return self._name_f

    @property
    def op(self) -> Op:
        return self._op

    @property
    def channel(self) -> Channel:
        return self._channel


#


class RulesCollection(lang.Final, ta.Collection[Rule]):
    def __init__(self, *rules: ta.Union[Rule, 'RulesCollection']) -> None:
        super().__init__()

        rules_set: set[Rule] = set()
        rules_by_name: dict[str, Rule] = {}
        rules_by_name_f: dict[str, Rule] = {}
        rules_by_op: dict[Op, Rule] = {}

        def add(gr: Rule) -> None:
            check.isinstance(gr, Rule)

            check.not_in(gr, rules_set)
            check.not_in(gr._name, rules_by_name)  # noqa
            check.not_in(gr._name_f, rules_by_name_f)  # noqa
            check.not_in(gr._op, rules_by_op)  # noqa

            rules_set.add(gr)
            rules_by_name[gr._name] = gr  # noqa
            rules_by_name_f[gr._name_f] = gr  # noqa
            rules_by_op[gr._op] = gr  # noqa

        for e in rules:
            if isinstance(e, RulesCollection):
                for c in e:
                    add(c)
            else:
                add(e)

        self._rules_set = rules_set
        self._rules_by_name: ta.Mapping[str, Rule] = rules_by_name
        self._rules_by_name_f: ta.Mapping[str, Rule] = rules_by_name_f
        self._rules_by_op: ta.Mapping[Op, Rule] = rules_by_op

    @property
    def rules_set(self) -> ta.AbstractSet[Rule]:
        return self._rules_set

    @property
    def rules_by_name(self) -> ta.Mapping[str, Rule]:
        return self._rules_by_name

    @property
    def rules_by_name_f(self) -> ta.Mapping[str, Rule]:
        return self._rules_by_name_f

    @property
    def rules_by_op(self) -> ta.Mapping[Op, Rule]:
        return self._rules_by_op

    #

    def __len__(self) -> int:
        return len(self._rules_set)

    def __iter__(self) -> ta.Iterator[Rule]:
        return iter(self._rules_set)

    def __contains__(self, item: Rule) -> bool:  # type: ignore[override]
        return item in self._rules_set

    #

    def rule(self, name: str) -> Rule | None:
        return self._rules_by_name_f.get(name.casefold())


##


class Grammar(lang.Final):
    def __init__(
            self,
            *rules: Rule | RulesCollection,
            root: Rule | str | None = None,
    ) -> None:
        super().__init__()

        if len(rules) == 1 and isinstance(r0 := rules[0], RulesCollection):
            self._rules = r0
        else:
            self._rules = RulesCollection(*rules)

        if isinstance(root, str):
            root = self._rules.rules_by_name_f[root.casefold()]
        self._root = root

    @property
    def rules(self) -> RulesCollection:
        return self._rules

    @property
    def root(self) -> Rule | None:
        return self._root

    #

    def rule(self, name: str) -> Rule | None:
        return self._rules.rule(name)

    def replace_rules(self, *rules: Rule) -> 'Grammar':
        rc = RulesCollection(*rules)
        if rc.rules_set == self._rules.rules_set:
            return self

        return Grammar(
            rc,
            root=self._root.name if self._root is not None else None,
        )

    #

    def iter_parse(
            self,
            source: str,
            root: Rule | str | None = None,
            *,
            start: int = 0,
            debug: int = 0,
            **kwargs: ta.Any,
    ) -> ta.Iterator[Match]:
        if root is None:
            if (root := self._root) is None:
                raise AbnfError('No root or default root specified')
        else:
            if isinstance(root, str):
                root = self._rules.rules_by_name_f[root.casefold()]
            else:
                root = check.in_(check.isinstance(root, Rule), self._rules)

        parser = parsing._make_parser(  # noqa
            self,
            source,
            debug=debug,
            **kwargs,
        )

        try:
            yield from parser.iter_parse(
                root._op,  # noqa
                start,
            )
        finally:
            pass  # Debugging aid to inspect parser state after parse

    def parse(
            self,
            source: str,
            root: str | None = None,
            *,
            start: int = 0,
            complete: bool = False,
            debug: int = 0,
            **kwargs: ta.Any,
    ) -> Match | None:
        if (match := longest_match(self.iter_parse(
                source,
                root,
                start=start,
                debug=debug,
                **kwargs,
        ))) is None:
            return None

        if complete and (match.start, match.end) != (start, len(source)):
            raise AbnfIncompleteParseError

        return match



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from ...lite.namespaces import NamespaceClass
from .core import IoPipelineMessages


T = ta.TypeVar('T')


##


class AsyncIoPipelineMessages(NamespaceClass):
    @ta.final
    @dc.dataclass(frozen=True)
    class Await(
        IoPipelineMessages.Completable[T],
        IoPipelineMessages.NeverInbound,
        ta.Generic[T],
    ):
        obj: ta.Awaitable[T]



import typing as ta

import anyio

from ..identities import AsyncliteIdentities


##


class AnyioAsyncliteIdentities(AsyncliteIdentities):
    def current_identity(self) -> ta.Any:
        return anyio.get_current_task()



# ruff: noqa: UP006 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta


T = ta.TypeVar('T')


##


@dc.dataclass()
class MinUniquePrefixNode(ta.Generic[T]):
    part: ta.Sequence[T] = ()

    children: ta.Dict[T, 'MinUniquePrefixNode[T]'] = dc.field(default_factory=dict)
    count: int = 0  # number of items in this subtree

    terminals: ta.Set[ta.Sequence[T]] = dc.field(default_factory=set)
    terminal_items: ta.Dict[ta.Sequence[T], ta.Sequence[T]] = dc.field(default_factory=dict)

    min_unique_prefix_len: int = dc.field(init=False)

    @property
    def terminal_count(self) -> int:
        return len(self.terminals)

    @dc.dataclass()
    class NonUniqueKeyError(Exception):
        key: ta.Any

    def lookup(self, prefix: ta.Sequence[T]) -> ta.Sequence[T]:
        try:
            return self._lookup(prefix)
        except MinUniquePrefixNode._NonUniqueKeyError:
            raise MinUniquePrefixNode.NonUniqueKeyError(prefix)  # noqa

    def __len__(self) -> int:
        return self.count

    def __iter__(self) -> ta.Iterator[ta.Sequence[T]]:
        yield from self.terminal_items.values()
        for child in self.children.values():
            yield from child

    #

    @staticmethod
    def _common_prefix_len(
            item: ta.Sequence[T],
            item_pos: int,
            part: ta.Sequence[T],
    ) -> int:
        n = min(len(item) - item_pos, len(part))
        i = 0
        while i < n and item[item_pos + i] == part[i]:
            i += 1
        return i

    def _match_child(
            self,
            item: ta.Sequence[T],
            item_pos: int,
    ) -> ta.Tuple[ta.Optional['MinUniquePrefixNode[T]'], int]:
        if item_pos == len(item):
            return None, 0

        child = self.children.get(item[item_pos])
        if child is None:
            return None, 0

        return child, self._common_prefix_len(item, item_pos, child.part)

    def _add_terminal(
            self,
            term: ta.Sequence[T],
            item: ta.Sequence[T],
    ) -> None:
        if term in self.terminal_items:
            raise ValueError('duplicate items present')

        self.terminals.add(term)
        self.terminal_items[term] = item

    def _set_min_unique_prefix_len(self, depth_before: int, *, is_root: bool) -> int:
        if self is is_root and self.count <= 1:
            ret = 0

        elif self.count == 1:
            ret = depth_before + 1

        else:
            child_depth_before = depth_before + len(self.part)
            ret = max(
                child._set_min_unique_prefix_len(child_depth_before, is_root=False)  # noqa
                for child in self.children.values()
            )

        self.min_unique_prefix_len = ret
        return ret

    class _NonUniqueKeyError(Exception):
        pass

    def _find_unique_terminal_item(self: 'MinUniquePrefixNode[T]') -> ta.Sequence[T]:
        cur = self

        while True:
            if cur.count != 1:
                raise MinUniquePrefixNode._NonUniqueKeyError

            if cur.terminal_items:
                if len(cur.terminal_items) != 1:
                    raise RuntimeError('invalid tree')
                return next(iter(cur.terminal_items.values()))

            if not cur.children:
                raise RuntimeError('invalid tree')

            if len(cur.children) != 1:
                raise RuntimeError('invalid tree')

            cur = next(iter(cur.children.values()))

    def _lookup(self, prefix: ta.Sequence[T]) -> ta.Sequence[T]:
        node = self
        pos = 0

        while True:
            common_len = self._common_prefix_len(prefix, pos, node.part)
            needed = min(len(prefix) - pos, len(node.part))
            if common_len != needed:
                raise KeyError(prefix)

            pos += common_len
            if pos == len(prefix):
                return node._find_unique_terminal_item()  # noqa

            child, _ = node._match_child(prefix, pos)  # noqa
            if child is None:
                raise KeyError(prefix)

            node = child


def build_min_unique_prefix_tree(
        items: ta.Sequence[ta.Sequence[T]],
        *,
        subsequence: ta.Optional[ta.Callable[[ta.Sequence[T]], ta.Sequence[T]]] = None,
) -> MinUniquePrefixNode[T]:
    if subsequence is None:
        subsequence = lambda x: x  # noqa

    root: MinUniquePrefixNode[T]

    if not items:
        root = MinUniquePrefixNode()
        root.min_unique_prefix_len = 0
        return root

    if len(items) == 1:
        part = subsequence(items[0])
        root = MinUniquePrefixNode(
            part=part,
            count=1,
        )
        root._add_terminal(part, items[0])  # noqa
        root.min_unique_prefix_len = 0
        return root

    root = MinUniquePrefixNode()

    for item in items:
        root.count += 1

        node = root
        pos = 0

        while True:
            if pos == len(item):
                node._add_terminal(node.part, item)  # noqa
                break

            child, common_len = node._match_child(item, pos)  # noqa
            if child is None:
                part = subsequence(item[pos:])
                new_child = MinUniquePrefixNode(
                    part=part,
                    count=1,
                )
                new_child._add_terminal(part, item)  # noqa
                node.children[part[0]] = new_child
                break

            if common_len == len(child.part):
                child.count += 1
                node = child
                pos += common_len
                continue

            if pos + common_len == len(item):
                child.count += 1
                child._add_terminal(child.part[:common_len], item)  # noqa
                break

            split = MinUniquePrefixNode(
                part=child.part[:common_len],
                count=child.count + 1,
            )
            node.children[item[pos]] = split

            old_terminals = child.terminals
            old_terminal_items = child.terminal_items

            child.terminals = set()
            child.terminal_items = {}

            child.part = child.part[common_len:]

            for term in old_terminals:
                old_item = old_terminal_items[term]
                if len(term) <= common_len:
                    split._add_terminal(term, old_item)  # noqa
                else:
                    child._add_terminal(term[common_len:], old_item)  # noqa

            split.children[child.part[0]] = child

            pos += common_len
            new_part = subsequence(item[pos:])
            new_child = MinUniquePrefixNode(
                part=new_part,
                count=1,
            )
            new_child._add_terminal(new_part, item)  # noqa
            split.children[new_part[0]] = new_child

            break

    stack = [root]
    while stack:
        node = stack.pop()

        if len(node.terminals) > 1:
            raise ValueError('at least one item is a prefix of another')

        if node.terminals:
            term = next(iter(node.terminals))
            if len(term) < len(node.part):
                raise ValueError('at least one item is a prefix of another')
            if node.children:
                raise ValueError('at least one item is a prefix of another')

        stack.extend(node.children.values())

    root._set_min_unique_prefix_len(0, is_root=True)  # noqa

    return root


##


def min_unique_prefix_lens(
        items: ta.Sequence[ta.Sequence[T]],
        *,
        subsequence: ta.Optional[ta.Callable[[ta.Sequence[T]], ta.Sequence[T]]] = None,
) -> ta.List[int]:
    if len(items) <= 1:
        return [0] * len(items)

    root = build_min_unique_prefix_tree(
        items,
        subsequence=subsequence,
    )

    out = []

    for item in items:
        node = root
        pos = 0
        depth = 0

        while True:
            child = node.children[item[pos]]

            if child.count == 1:
                out.append(depth + 1)
                break

            part_len = len(child.part)
            pos += part_len
            depth += part_len
            node = child

    return out


def min_unique_prefix_len(
        items: ta.Sequence[ta.Sequence[T]],
        *,
        subsequence: ta.Optional[ta.Callable[[ta.Sequence[T]], ta.Sequence[T]]] = None,
) -> int:
    if len(items) <= 1:
        return 0

    root = build_min_unique_prefix_tree(
        items,
        subsequence=subsequence,
    )

    return root.min_unique_prefix_len



import typing as ta
import warnings

from .. import check
from .. import lang
from .debug import _ResourcesDebug
from .errors import ResourceNotEnteredError
from .errors import UnclosedResourceWarning


##


class BaseSimpleResource(_ResourcesDebug, lang.Abstract):
    """
    IMPORTANT: Subclasses must not raise exceptions in constructors or resources will be leaked (given the intended
    usage ergonomics). Do any validation in `_enter`.
    """

    _resource_state: ta.Literal[
        'new',

        'entering',
        'failed_entering',
        'entered',

        'closing',
        'failed_closing',
        'closed',
    ] = 'new'

    #

    @ta.final
    @property
    def _entered(self) -> bool:
        return self._resource_state == 'entered'

    @ta.final
    @property
    def _closed(self) -> bool:
        return self._resource_state in ('failed_closing', 'closed')

    #

    @ta.final
    def _check_entered(self) -> None:
        if not self._entered:
            raise ResourceNotEnteredError(self)  # noqa

    #

    @property
    def _is_resourceless(self) -> bool:
        return False

    @ta.final
    def __del__(self) -> None:
        if (
                self._entered and
                not self._closed and
                not self._is_resourceless and
                self._resources_debug
        ):
            warnings.warn(
                f'\n\n{(sep := ("=" * 40))}\n'
                f'{self.__class__.__name__} object '
                f'{self._resources_debug_repr or f"at {hex(id(self))}"} '
                f'was not properly closed before finalization. Please ensure that `__exit__` or `close()` is called '
                f'before the object is finalized.'
                f'\n\n'
                f'{"".join(self._resources_debug_traceback).rstrip()}'
                f'\n{sep}\n',
                UnclosedResourceWarning,
            )


class SimpleResource(BaseSimpleResource, lang.Abstract):
    @ta.final
    def __enter__(self) -> ta.Self:
        check.state(self._resource_state == 'new')
        self._resource_state = 'entering'
        try:
            self._enter()
            self._init_debug()
        except BaseException as e:  # noqa
            self._resource_state = 'failed_entering'
            raise
        self._resource_state = 'entered'
        return self

    def _enter(self) -> None:
        pass

    #

    @ta.final
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._close_internal(exc_val)

    @ta.final
    def close(self) -> None:
        self._close_internal(None)

    @ta.final
    def _close_internal(self, reason: BaseException | None) -> None:
        if not self._closed:
            try:
                self._close(reason)
            except BaseException as e:  # noqa
                self._resource_state = 'failed_closing'
                raise
            self._resource_state = 'closed'

    def _close(self, reason: BaseException | None) -> None:
        pass


class AsyncSimpleResource(BaseSimpleResource, lang.Abstract):
    @ta.final
    async def __aenter__(self) -> ta.Self:
        check.state(self._resource_state == 'new')
        self._resource_state = 'entering'
        try:
            await self._enter()
            self._init_debug()
        except BaseException as e:  # noqa
            self._resource_state = 'failed_entering'
            raise
        self._resource_state = 'entered'
        return self

    async def _enter(self) -> None:
        pass

    #

    @ta.final
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._close_internal(exc_val)

    @ta.final
    async def aclose(self) -> None:
        await self._close_internal(None)

    @ta.final
    async def _close_internal(self, reason: BaseException | None) -> None:
        if not self._closed:
            try:
                await self._close(reason)
            except BaseException as e:  # noqa
                self._resource_state = 'failed_closing'
                raise
            self._resource_state = 'closed'

    async def _close(self, reason: BaseException | None) -> None:
        pass



import abc
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from ... import reflect as rfl
from ..api.contexts import MarshalContext
from ..api.contexts import MarshalFactoryContext
from ..api.types import Marshaler
from ..api.types import MarshalerFactory
from ..api.values import Value
from .api import FieldTypeTagging
from .api import Impls
from .api import Polymorphism
from .api import TypeTagging
from .api import WrapperTypeTagging
from .impls import get_polymorphism_impls


##


class PolymorphismMarshaler(Marshaler, lang.Abstract):
    @abc.abstractmethod
    def get_impls(self) -> ta.Mapping[type, tuple[str, Marshaler]]:
        raise NotImplementedError


@dc.dataclass(frozen=True)
class WrapperPolymorphismMarshaler(PolymorphismMarshaler):
    m: ta.Mapping[type, tuple[str, Marshaler]]

    def get_impls(self) -> ta.Mapping[type, tuple[str, Marshaler]]:
        return self.m

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        tag, m = self.m[type(o)]
        return {tag: m.marshal(ctx, o)}


@dc.dataclass(frozen=True)
class FieldPolymorphismMarshaler(PolymorphismMarshaler):
    m: ta.Mapping[type, tuple[str, Marshaler]]
    tf: str

    def get_impls(self) -> ta.Mapping[type, tuple[str, Marshaler]]:
        return self.m

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        tag, m = self.m[type(o)]
        return {self.tf: tag, **m.marshal(ctx, o)}  # type: ignore


def make_polymorphism_marshaler(
        impls: Impls,
        tt: TypeTagging,
        ctx: MarshalFactoryContext,
) -> Marshaler:
    check.not_empty(impls)

    m = {
        i.ty: (i.tag, ctx.make_marshaler(i.ty))
        for i in impls
    }

    if isinstance(tt, WrapperTypeTagging):
        return WrapperPolymorphismMarshaler(m)
    elif isinstance(tt, FieldTypeTagging):
        return FieldPolymorphismMarshaler(m, tt.field)
    else:
        raise TypeError(tt)


@dc.dataclass(frozen=True)
class PolymorphismMarshalerFactory(MarshalerFactory):
    p: Polymorphism
    tt: TypeTagging = WrapperTypeTagging()

    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if (impls := get_polymorphism_impls(rty, self.p)) is None:
            return None
        return lambda: make_polymorphism_marshaler(impls, self.tt, ctx)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - cascading?
 - caching?
 - purging?
 - wishlist: piecewise config parts, unmarshal, registered via manifests
"""
import abc
import os.path
import typing as ta

from ..lite.abstract import Abstract
from ..lite.check import check
from ..os.shadow import ShadowPaths
from .formats import DEFAULT_CONFIG_FILE_LOADER
from .formats import ConfigFileLoader


ShadowConfig = ta.Mapping[str, ta.Any]  # ta.TypeAlias


##


class ShadowConfigs(Abstract):
    @abc.abstractmethod
    def get_shadow_config(self, path: str, name: str) -> ta.Optional[ShadowConfig]:
        raise NotImplementedError


class FileShadowConfigs(ShadowConfigs, Abstract):
    @abc.abstractmethod
    def get_shadow_config_file_path(
            self,
            path: str,
            name: str,
            *,
            preferred_ext: ta.Optional[str] = None,
    ) -> ta.Optional[str]:
        raise NotImplementedError


##


class NopShadowConfigs(ShadowConfigs):
    def get_shadow_config(self, path: str, name: str) -> ta.Optional[ShadowConfig]:
        return None


##


class FileShadowConfigsImpl(FileShadowConfigs):
    def __init__(
            self,
            paths: ShadowPaths,
            subdir: ta.Optional[str] = None,
            *,
            loader: ta.Optional[ConfigFileLoader] = None,
            preferred_ext: ta.Optional[str] = None,
    ) -> None:  # noqa
        super().__init__()

        if loader is None:
            loader = DEFAULT_CONFIG_FILE_LOADER

        if preferred_ext is not None:
            check.arg(not preferred_ext.startswith('.'))
            check.non_empty_str(preferred_ext)
            check.in_(preferred_ext, loader.file_exts)

        self._paths = paths
        self._subdir = subdir
        self._loader = loader
        self._preferred_ext = preferred_ext

    def get_shadow_config_file_path(
            self,
            path: str,
            name: str,
            *,
            preferred_ext: ta.Optional[str] = None,
    ) -> ta.Optional[str]:
        check.non_empty_str(name)
        check.not_in(os.sep, name)

        shadow = self._paths.get_shadow_path(path)
        if (sd := self._subdir):
            shadow = os.path.join(shadow, sd)

        if (found := self._loader.find_file(os.path.join(shadow, name))) is not None:
            return found
        elif (px := (preferred_ext if preferred_ext is not None else self._preferred_ext)) is not None:
            return os.path.join(shadow, f'{name}.{px}')
        else:
            return None

    def get_shadow_config(self, path: str, name: str) -> ta.Optional[ShadowConfig]:
        if (file_path := self.get_shadow_config_file_path(path, name)) is None:
            return None

        if not os.path.exists(file_path):
            return None

        return self._loader.load_file(file_path).as_map()
